#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <time.h>
#include <string.h>
#include <dirent.h>
#include "qheader.h"

#ifndef QDOS
#define SEP_CHAR '.'
#define DIR_SEP "/"
# define FNSIZE 256
#else
#include <stddef.h>
#include <qdos.h>
#include <sms.h>
# include <things.h>
# define STUFF_SIZE 10
# define TEST_ID  0x302e3031	/* 0.01 */
# define SEP_CHAR '_'
# define DIR_SEP "_"
# define FNSIZE 40
#define ADDTHING 1
static char *Stuff = "QFAX Items";
char _prog_name[] = "LoadNT";
void (*_SigStart)(void) = NULL;

#endif

#ifdef UQLX
#define LAST_CHAR '.'
#else
#define LAST_CHAR SEP_CHAR
#endif

typedef union 
{
    char txt[4];
    long x;
} ALIGN;

typedef struct
{
    char path[FNSIZE];
    time_t stamp;
} FAXITEM;

typedef struct
{
    int  flag;
    int  type;
    short maxitems;
    short current;
    short nitems;
    FAXITEM fdata[1];
} FAX_LIST;



int cmpf(const void *v1, const void *v2)
{
    return((*(FAXITEM **)v2)->stamp - (*(FAXITEM **)v1)->stamp);
}

#ifdef ADDTHING
static ALIGN thh = {THH_FLAG};
static void * thg_malloc(long n, jobid_t j)
{
    void *p;
    long dummy;
    p = sms_achp (n, &dummy, j);
}

static void * thg_free(void *p)
{
    sms_rchp (p);
}			     

int AddThing (short items, THING_LINKAGE ** p)
{
    int m, n;
    int size;
    int sts;


    size = (offsetof (THING_LINKAGE, th_name_text) + STUFF_SIZE );

    m = size + sizeof (FAX_LIST) + (items - 1) * sizeof (FAXITEM);
    *p = (THING_LINKAGE *) thg_malloc (m, 0);

    if (*p)
    {
	(*p)->th_name = STUFF_SIZE;
	memcpy ((*p)->th_name_text, Stuff, STUFF_SIZE);
	(*p)->th_verid = TEST_ID;

	(*p)->th_thing = ((char *) (*p) + size);

	((FAX_LIST *) (*p)->th_thing)->flag = thh.x;
	((FAX_LIST *) (*p)->th_thing)->type = THT_DATA;

	((FAX_LIST *) (*p)->th_thing)->maxitems = items;
	((FAX_LIST *) (*p)->th_thing)->current = 0;
	((FAX_LIST *) (*p)->th_thing)->nitems = 0;

	sts = sms_lthg (*p);
	if (sts != 0)
	{
	    thg_free (*p);
	    *p = NULL;
	}
    }
    return sts;
}
int AddFaxItem (FAX_LIST * f, FAXITEM *fi)
{
    short pos;

    _super ();
    pos = (f->current + 1) % f->maxitems;
    if (f->nitems < f->maxitems)
    {
	f->nitems++;
    }
    strcpy (f->fdata[f->current].path, fi->path);
    f->fdata[f->current].stamp = TIME_UNIX_QL(fi->stamp);
    f->current = pos;
    _user ();
}
#endif

ProcessDir(char *nam, int *pac, FAXITEM **pav, int *pndir, int *pnfile)
{
    DIR *d;
    struct stat s;
    int n; 

    fputs("Checking ... ", stdout);
    puts(nam);
    d = opendir(nam);
    if(d)
    {
        struct dirent *p;
        (*pndir)++;
        while(p = readdir(d))
        {
            char d_name[FNSIZE];
	    
	    if(*p->d_name == '.') continue;
	    
            strcpy(d_name, nam);
            strcat(d_name, p->d_name);
	    
            n = stat(d_name, &s);
            if(S_ISDIR(s.st_mode))
            {
		if(*(d_name+strlen(d_name)-1) != SEP_CHAR)
		    strcat(d_name, DIR_SEP);
                ProcessDir(d_name, pac, pav, pndir, pnfile);
            }
            else
            {
                char *q;
                (*pnfile)++;

                if(q = strrchr(d_name, LAST_CHAR))
                {
                    if(strcasecmp(++q, "fax") == 0)
                    {
                        FAXITEM *f;
                        if(f = malloc(sizeof(FAXITEM)))
                        {
                            strcpy(f->path, d_name);
                            f->stamp = s.st_ctime;
                            pav[(*pac)++] = f;
                        }
                    }
                }
            }
        }
    }     
    closedir(d);
}

main(int ac, char **av)
{
    FAXITEM **xav;
    int xac = 0, ndir = 0, nfile = 0, nfax;
    clock_t t1,t2;
    
#ifdef QDOS
    int sts;
    THING_LINKAGE *p = NULL;
    long vers;
    t1 = time(NULL);
#else
    FILE *fp;
    short maxac;
    t1 = clock();
#endif

#ifdef ADDTHING
    sts = (int) sms_uthg (Stuff, -1, -1, NULL, NULL, &vers, &p);
    if (sts != ERR_NF)
    {
        puts("Thing exists");
        exit(0);
    }
#endif

    xac = 0;
    xav = (FAXITEM **) malloc(2048*sizeof(FAXITEM *));
    ProcessDir(av[1], &xac, xav, &ndir, &nfile);
    nfax = xac;
    if(xac > 0)
    {
        puts("Sorting ... ");    
        qsort(xav, xac, sizeof(FAXITEM *), cmpf);

        if(xac > 20)
        {
            xac = 20;
        }
#ifdef ADDTHING
        AddThing(xac, &p);
#endif
#ifndef QDOS
        maxac = 0;
        fp = fopen("qfax/QFAXThing", "w");
#endif
        while(xac--)
        {
            FAXITEM *f;
            f = xav[xac];
#ifdef QDOS

#ifdef ADDTHING
            AddFaxItem((FAX_LIST *) p->th_thing, f);
#endif
#else

            fwrite(f, 1, sizeof(FAXITEM), fp);
#endif
            printf("%2d: %-40.40s %s", xac, f->path, ctime(&(f->stamp)));
        }
    }
#ifdef QDOS
    t2 = time(NULL) - t1;
    printf("Found %d fax from %d files in %d directories in %d seconds\n",
        nfax, nfile, ndir, t2);            
#else
    t2 = clock() - t1;
    printf("Found %d fax from %d files in %d directories in %d.%02d seconds\n",
        nfax, nfile, ndir, t2/50, (t2 % 50)*2);            
    fclose (fp);
#endif
    exit(0);
}    
