#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <stdarg.h>
#include <limits.h>
#include <sys/types.h>
#ifdef __linux__
# include <sys/wait.h>
#endif
#ifdef _DCC
 #include <getopt.h>
 #include <clib/dos_protos.h>
 extern void GiveItBack (void);
#endif

#ifdef __NT__
#include <getopt.h>
extern int optind;
extern char *optarg;
#endif

#include "qfaxparse.h"
#include "qheader.h"
#include "qfv.h"
#include "qfpp.h"

#ifdef QDOS
 #include <memory.h>
 #include <qdos.h>
 #define SEPCHAR '_'
 void InitWindow (short, short);
 void MyExit (short, int);
 long _stack = 8400;
#else
# define SEPCHAR '.'
# define MyExit(a,b) exit(b)
# ifndef __NT__
  pid_t pid = 0;
# endif
#endif

static VIEWPORT vp = {0,};
PVIEWPORT pv = &vp;

int decode_fax2 (int *, int, short, short *);
int FindFax (short);
int PrintPage (int, char *);
void ShowHelp (void);
#ifndef __NT__
pid_t preview (char *, char *);
void WaitPid (pid_t);
#endif
void IniPrt(FAXDATA *);
void ScanFaxFile(void);

FAXDATA qd = {0};

extern void CloseFaxFile (void);
QFVREC *qpr = NULL;
static char * range = NULL;

void CleanUp (void)
{
    static short na = 0;

    if (!na && qpr == NULL)
    {
	na = 1;
#ifdef _DCC
	GiveItBack ();
#endif
	CloseFaxFile ();
#if !defined (__linux__) && !defined (__NT__)
	if ((pv->interact & 2) && pv->preq)
	{
	    ClearRequester (pv->preq);
	}
#endif
    }
}

int GetPageMap (void)
{
    short actual;
    int bitsize;

    if (pv->qh.version > QFAXNEWFMT && pv->qh.hint)
    {
	actual = (SLACK + (pv->qh.hint << (1 - pv->qh.resolution)));
	if (actual > pixhgt[pv->qh.page])
	{
	    actual = pixhgt[pv->qh.page];
	}
    }
    else
    {
	actual = pixhgt[pv->qh.page];
    }

    bitsize = (pixwid[pv->maxwid] * actual) >> 3;
    pv->scan = malloc (bitsize);
    pv->buck = actual-6;
    ShowReport(&qd, REP_PRT, "Buck stops at %d\n", pv->buck);
    pv->maxmem = bitsize;
    return bitsize;
}


#ifdef __linux__
void strins (char *d, const char *s)
{
    int len = strlen (s);	/*  # bytes to insert   */
    char *ptr;

    /*
     *        make room
     */

    ptr = d + strlen (d);
    while (ptr >= d)
    {
	ptr[len] = ptr[0];
	--ptr;
    }

    /*
     *        insert string
     */

    while (*s)
	*++ptr = *s++;
}
#endif

int DoPage (short skipbad)
{
    int res = -3;
    short badlines = 0;
    int ip = 0;
    
    if (*pv->FileName && FindFax (2))
    {
	if(range) ScanFaxFile();
	
	if (pv->xpl && pv->qh.width)
	{
	    Report ("Unsupported width");
	}
	else
	{
	    short ipg = -1;
	    int bitmem;

	    bitmem = GetPageMap ();

	    while (pv->scan == NULL && pv->qh.page)
	    {
		ipg = --(pv->qh.page);
		bitmem = GetPageMap ();
	    }
	    if (pv->scan)
	    {
		int i = 0;

		if (ipg != -1)
		{
		    CloseFaxFile ();
		    FindFax (ipg);
		}

		Report ("Using page size %d, (%d bytes), %d%c",
			pv->qh.page, bitmem,
			pv->qh.coding + 1,
			pv->qh.resolution ? 'h' : 'l');

		Report ("Decoding %s ... %d (real) pages", pv->FileName,
			pv->qh.pagemax);
#if !defined (__linux__) && !defined (__NT__)
		if (pv->interact & 2)
		{
		    pv->preq = SetRequester (NULL, pv->FileName);
		}
#endif
		do
		{
		    char tmp[PATH_MAX];
		    PRVIEW *prv;
#ifndef __NT__
		    pid_t pid;
#endif
 		    int maxl;
		    
		    for (prv = qd.prview; prv; prv = prv->next)
		    {
			if (prv->value == pv->xpl->type)
			{
			    break;
			}
		    }

		    if(!range || *(range + i))
		    {
			memset (pv->scan, 0, bitmem);
			
			res = decode_fax2 (&maxl, i, skipbad, &badlines);
			
			if (res == 0 && pv->pabort == 0)
			{
			    Report ("%3d: Scan lines %d/%d (%d bad)", i + 1,
				    pv->xline, maxl,badlines);
			    if (pv->xpl->paged)
			    {
				sprintf (tmp,
					 (*pv->PrtName) ? pv->PrtName : "pfax%03d%c%s",
					 i + 1, SEPCHAR, pv->xpl->name);
				pv->fp = fopen (tmp, "wb");
			    }
			    PrintPage (ip, tmp);
			    ip++;
			    
			    if (pv->xpl->paged && pv->fp)
			    {
				fclose (pv->fp);
#ifndef __NT__
				if (i == 0 && prv && prv->program)
				{
				    pid = preview (prv->program, tmp);
				}
#endif
			    }
			}
			else
			{
			    Report ("Decode fails %d", res);
			}
			res = (pv->pabort == 2);
		    }
		    else res = 0;
		    
		    i++;
		}
		while (i < pv->qh.pagemax && res == 0);
	    }
#if !defined (__linux__) && !defined(__NT__)
	    if ((pv->interact & 2) && pv->preq)
	    {
		ClearRequester (pv->preq);
		pv->preq = NULL;
	    }
#endif
	}
    }
    free (pv->scan);
    pv->scan = NULL;
    return res;
}

#ifdef QDOS
int (*_cmdwildcard) (char *, char ***, int *) = cmdexpand;

#else
# ifdef _DCC
extern int WildCards (int *, char ***);

# endif
#endif

void addrange(char *buf, char *range, int nn)
{
    char *p,*q;
    short j,n,ln = 0;
    
    for(p = buf; *p && (n = strtol(p, &q, 0)); p = q+1)
    {
	switch(*q)
	{
   	  case ':':
	  case '-':
	    ln = n;
	    break;
	    
	  case ',':
	  case '\0':
	    if(ln)
	    {
		for(j = ln; j <= n; j++)
		{
		    range[j-1]++;
		}
		ln = 0;
	    }
	    else
		range[n-1]++;
	    break;
	}
    }
}

short toskip = 0;
short skipall = 0;
extern short mineop;
	
int main (int ac, char **av)
{
    int c,i;
    int res = 0;
    short uselast = 0, pause = 0;
    void *mywin = NULL;
    short skipbad = 1;
    char repmsg[80];
    short del = 0;
#ifdef QDOS
    short wait = 1;
#endif    
    *repmsg = 0;
    
    atexit (CleanUp);
#ifdef _DCC
    WildCards (&ac, &av);
#endif

    pv->interact = isatty (1);
    for(i = 1; i < ac; i++)
    {
	char *p = NULL;

	if (**(av+i) == '-')
	{
	    if (*(1+*(av+i)) == 'C')
	    {
		if (*(2+*(av+i)))
		{
		    p = (2+*(av+i));
		}
		else if (*(av + 1 + i) && **(av + 1 + i) != '-')
		{
		    p = *(av+i+1);
		}
		if (p)
		{
		    char t[80];
		    if(access(p, R_OK) == 0)
		    {
			sprintf(t,"QFAX_DATA=%s",p);
			putenv(t);
		    }
		}
		break;
	    }
	}
    }

    FaxParams (&qd);
    pv->paper = qd.paper;

    if (qd.pfaxini)
    {
	FrigArgs (&ac, &av, qd.pfaxini);
    }

    pv->xpl = XLateXPL(0, NULL);
   
    while ((c = getopt (ac, av, "dbavqpt:o:xc:s:Z:f:l+:r:J:j:P:C:")) != -1)
    {
	switch (c)
	{
	  case 'J':
	    skipall = 1;
	  case 'j':
	    toskip = atoi(optarg);
	    break;
	  case 'P':
	    mineop = atoi(optarg);
	    break;
	  case 'r':
	    if(range == NULL)
	    {
		range = calloc(1,256);
	    }
	    addrange (optarg, range, 256);
	    break;
	  case '+':
	    qd.report ^= atoi(optarg);
	    break;
	  case 'd':
	    del = 1;
	    break;
	  case 'b':
	    skipbad = 0;
	    break;
	  case 'l':
	    uselast = 1;
	    break;
	  case 'f':
	    pv->paper = atoi (optarg);
	    break;
	  case 'a':
	    pv->interact |= 2;
	    break;
	  case 's':
	    pv->specific = atoi (optarg);
	    break;
	  case 'c':
	    pv->compress = atoi (optarg);
	    break;
	  case 'q':
	    pv->interact &= ~1;
	    break;
	  case 'v':
	    pv->interact |= 1;
	    break;
	  case 'p':
	    pause = 1;
	    break;
	  case 'o':
	    strcpy (pv->PrtName, optarg);
	    break;
	  case 't':
	    if(isdigit(*optarg))
	    {
		short t;
		
		t = atoi (optarg);
		if((pv->xpl = XLateXPL (t, NULL)))
		{
		    sprintf(repmsg, "-t %d is deprecated, use -t %s",
			    pv->xpl->type, pv->xpl->name);
		}
	    }
	    else
	    {
		pv->xpl = XLateXPL(0, optarg);
	    }
	    
	    if (!pv->xpl)
	    {
		res = 1;
	    }
	    break;
	  case 'x':
#ifdef QDOS
	    wait = 0;
#endif
	    break;
	    
	  case 'Z':
	    qpr = (QFVREC *) strtol (optarg, NULL, 16);
	    strcpy (pv->FileName, qpr->file);
	    pv->scan = qpr->scan;
	    pv->xline = qpr->xline;
	    pv->qh = qpr->qh;
	    pv->interact = 2;
	    mywin = qpr->win;
	    break;
	  case '?':
	    res = 1;
	    break;
	}
    }

    if(skipall) skipall = toskip;
    
    if (res)
    {
	ShowHelp ();
	MyExit (wait, -15);
    }

#ifdef QDOS
    {
	int qchd;
        if (pv->xpl->paged == 0 && *pv->PrtName != '|')
        {
            while ((qchd = io_open (pv->PrtName, 3)) == ERR_IU)
            {
                sleep (5);
            }
            io_close (qchd);
        }
	InitWindow (0, ((*qd.mono) & 1));
    }

#endif
    if (*pv->PrtName == '\0')
    {
	PRVIEW *prv;

	for (prv = qd.prview; prv; prv = prv->next)
	{
	    if (prv->value == pv->xpl->type)
	    {
		if (prv->file)
		{
		    strcpy (pv->PrtName, prv->file);
		}
	    }
	}
    }

    Report ("Using %s driver", pv->xpl->name);

    if(*repmsg)
    {
	Report(repmsg);
    }

#ifndef __NT__
    if (*pv->PrtName == '|')
    {
        pv->fp = popen (pv->PrtName + 1, "w");
        pv->xpl->paged = 0;
        
    }
    else
#endif
    if (pv->xpl->paged == 0)
    {
        if (*pv->PrtName == 0)
        {
            sprintf (pv->PrtName, "prtqfax%c%s", SEPCHAR, pv->xpl->name);
        }
        pv->fp = fopen (pv->PrtName, "wb");

	if (pv->fp == NULL)
	{
	    MyExit (wait, 0);
	}
	IniPrt(&qd);
    }
 
    pv->pabort = 0;
#ifdef _DCC
    if (qpr)
    {
	if (pv->interact & 2)
	{
	    pv->preq = SetRequester (mywin, pv->FileName);
	}
	PrintPage (0);
	if ((pv->interact & 2) && pv->preq)
	{
	    ClearRequester (pv->preq);
	}
    }
    else
#endif
    {
	if (uselast)
	{
	    FILE *fp;
	    char *p, qnam[80];

#ifndef _DCC
	    sprintf (qnam, "%sQFAX-LastFax", qd.in_dir);
#else
	    strcpy (qnam, qd.in_dir);
	    AddPart (qnam, "QFAX-LastFax", sizeof (qnam));
#endif
	    if ((fp = fopen (qnam, "r")))
	    {
		fgets (pv->FileName, sizeof (pv->FileName), fp);
		if ((p = strchr (pv->FileName, '\n')))
		{
		    *p = 0;
		}
		fclose (fp);
	    }
	    DoPage (skipbad);
	}
	else
	{
	    for (res = 0; res == 0 && *(av + optind); optind++)
	    {
		strcpy (pv->FileName, *(av + optind));
		res = DoPage (skipbad);
		if(del)
		{
		    unlink(pv->FileName);
		}
	    }
	}
    }

    ClosePrt ((pv->xpl->type == 0 && qd.paper != pv->paper));

    if (!pv->xpl->paged)
    {
#ifndef __NT__
	if (*pv->PrtName == '|')
	{
	    pclose (pv->fp);
	}
	else
#endif
	{
	    fclose (pv->fp);
	}
    }
   
    MyExit (wait, 0);
    return 0;
}
