#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <stdarg.h>
#include <time.h>

#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef __NT__
# include <sys/wait.h>
# ifndef O_BINARY
#  define O_BINARY 0
# endif
#else
char * stpcpy(char *d, const char *s);
#endif

#ifdef QDOS
# include <qdos.h>
#else
# ifdef _DCC
#  include <clib/dos_protos.h>
#  include <clib/exec_protos.h>
#  include <dos/dos.h>
#  include <dos/dostags.h>
#  include <devices/serial.h>
# endif
#endif

#include "qheader.h"
#include "qfaxparse.h"
#include "qfax.h"

#ifdef QDOS
#include <pd/csrvthg.h>
#include "qfaxsrv.h"
void AddFaxList (char *, time_t, char *);
extern int qsystem (int, char *, int);
short minnie = 0;
int OpenAppend (char *nam, int jobid);
#else
# ifdef _DCC
#  include <lib/rexx.h>
  extern void StartRexx (void);
# endif
#endif

FAXDATA qd = {0,};

typedef struct
{
    int vr;
    int br;
    int wd;
    int ln;
    int df;
    int ec;
    int bf;
    int st;
} T30PARAMS, *PT30PARAMS;

T30PARAMS t30p = {0};
T30PARAMS x30p = {0};
static T30PARAMS ini30p = {0};
static T30PARAMS l30p = {0};
static MHEAD PgH = {0};

#ifdef QDOS
  long _stack = 10240;
  char _prog_name[] = "qfax";
  char _version[] = FAXVERS;
# define SNDBUFSIZE 128 	/* write this many bytes at a time to fax */
#else
# define SNDBUFSIZE 512	        /* write this many bytes at a time to fax */
# include <getopt.h>
#endif

#define OK       'O'		/* first characters of modem prompts */
#define CONNECT  'C'
#define BUSY     'B'
#define RING     'R'

#define MAXIOPT   6		/* maximum # of modem initialization commands */

#define CMDBUFSIZE 256		/* longest possible modem response */

#define RINGPERIOD 8		/* maximum seconds between rings (usu. 2+4=6) */
#define RINGSTR    "RING"	/* modem RING response */
#define RINGSTRLEN 4		/* RING response length */

enum cchar
{				/* control characters */
    NUL, SOH, STX, ETX, EOT, ENQ, ACK, BEL, BS, HT, LF,
    VT, FF, CR, SO, SI, DLE, XON, DC2, XOFF, DC4, NAK,
    SYN, ETB, CAN, EM, SUB, ESC, FS, GS, RS, US
};

static char *ccname[32] =
{				/* control character names */
    "NUL", "SOH", "STX", "ETX", "EOT", "ENQ", "ACK", "BEL", "BS", "HT", "LF",
    "VT", "FF", "CR", "SO", "SI", "DLE", "XON", "DC2", "XOFF", "DC4", "NAK",
    "SYN", "ETB", "CAN", "EM", "SUB", "ESC", "FS", "GS", "RS", "US"};

static char *promptstr[] =
{				/* modem responses that are prompts */
    "OK", "CONNECT", "BUSY", "NO ", "ERROR", 0};

int hsc, good, ppr, poll = 0, ptsi = 0, ipoll = 0, fline = 0;
int online = 0;
char tsi[64];
time_t faxtime = 0;
static char ttsi[64];
static short alias = 0;
static short achar = 1;
static long eofPos = 0;
static short bopt = -1;
static short anon = 1;
static short isanon;

#ifdef QDOS
 static short domon = 0;
 static int poppid = 0;
 static int popres = -1;
 static int ctx = 0;
 static void *srvmsg = NULL;
 static short srvlen;
 static short Iserve = 0;
 static short dact = 0;
 char *ExitParams = NULL;
 extern jcall (void (*)(), char *, int, int, int,...);
 extern void answer (int *, int *);
# define QX_ERR_1   ERR_IU
# define QX_ERR_2   ERR_FE
# define QX_ERR_3   ERR_TE
# define QX_ERR_4   ERR_BP
# define QX_ERR_5   ERR_NF
# define QX_ERR_6   ERR_NI
#else
# define QX_ERR_1   1
# define QX_ERR_2   2
# define QX_ERR_3   3
# define QX_ERR_4   4
# define QX_ERR_5   5
# define QX_ERR_6   6
# define QX_ERR_255 255
#endif

#define QX_OK       0

#define QX_ERR_SEND 255
#define QX_ERR_GRAB 254
#define QX_FORCE    253
#define QX_DONE     252

#define ERR_CTRL_C      -1
#define ERR_TIMEOUT     -2

#define QREXX_EXIT      1
#define QREXX_SUSPEND   2
#define QREXX_WAKE      4

static short faa = 0;
static short act = 0;

FILE *LOG = NULL;
FILE *STDE = NULL;

#define FX_PREINI   0
#define FX_FCLASS   1
#define FX_FAA      2
#define FX_FLID     3
#define FX_FDCC     4
#define FX_USERAT   5

char dl_1[2][64] = {
    "+FCLASS=2",
    "+FCLASS=2.0"};
char dl_2[2][64] = {
    "+FBOR=0 +FCR=1 +FAA=0",
    "+FBO=0 +FCR=1 +FLO=2 +FAA=0 +FNR=1,1,1,0" 
};

char *fdcc[] = {"+FDCC", "+FCC"};
char *fk [] = {"+FK","+FSK"};
#define FBORLEN 6
char *fbor[] = {"+FBOR","+FBO"};
char *fspl0[] = {"+FSPL=0","+FSP=0"};
char *fspl1[] = {"+FSPL=1","+FSP=1"};
char *flpl1[] = {"+FLPL=1","+FLP=1"};
char *flid[] = {"+FLID", "+FLI"};
char *c2ppm[] = {"\020,","\020;","\020."};

char *iopt[MAXIOPT] = {NULL};

static char *dialcmd;
static char *inicmd;
static short prt = 0;
static char *qmode = "";
static short force = 0;

short c20 = 0;

unsigned char reversebits[256];
char LastNam[512];

#ifdef QDOS
# define Delay(p1) mt_susjb(-1, p1, NULL)
#endif

#if defined (__linux__) || defined (__NT__)
#ifdef __linux__
# define Delay(p1) usleep((p1)*20000)
#else
# define Delay(p1) Sleep((p1)*50)
#endif
extern FILE *lopen(const char *, const char *);
#else
#define lopen fopen
#endif

char *t30_vr_string (PT30PARAMS p);
char *t30_br_string (PT30PARAMS p);
char *t30_ln_string (PT30PARAMS p);
char *t30_wd_string (PT30PARAMS p);
char *t30_df_string (PT30PARAMS p);

typedef struct
{
    short rate;
    short res;
    short code;
    short page;
    short width;
    short lsb;
} MYPARAMS;

static MYPARAMS my = {0};

static void TidyUp (void);
static void GetQFax (void);
static short Grab = 0;

#define RTCLEN  9
#define RTCDAT  "\x00\x10\x01\x00\x10\x01\x00\x10\x01"

static char pollfile[128] = {0};

#ifdef __linux__
char *strlwr(char *p)
{
    char *q;
    for(q = p; *q; q++)
    {
	*q = tolower(*q);
    }
    return p;
}
#endif
static void AppendList (MHEAD * h, MLIST * m)
{
    if (h->head)
    {
	(h->tail)->next = m;
	h->tail = m;
    }
    else
    {
	h->head = h->tail = m;
    }
    m->next = NULL;
}

short msg (short code, char *s,...)
{

    time_t et;
    char *t, txt[132];
    struct tm *p;

    et = time (NULL);
    p = localtime (&et);

    if (LOG || STDE || code == QX_FORCE)
    {
	va_list ap;

	va_start (ap, s);

	sprintf (txt, "%02d:%02d:%02d ", p->tm_hour, p->tm_min, p->tm_sec);
	t = txt + strlen (txt);
	vsprintf (t, s, ap);
	va_end (ap);

	if (LOG)
	{
	    fputs (txt, LOG);
	    fputc ('\n', LOG);
	    fflush (LOG);
	}
	if (code == QX_FORCE)
	{
	    puts (txt);
	    fflush (stdout);
	}
	else if (STDE)
	{
	    fputs (txt, stderr);
	    fputc ('\n', stderr);
	    fflush (stderr);
	}
    }
    return code;
}

void logc (char c)
{
    char txt[8];

    if (LOG || STDE)
    {
	if (c >= ' ' && c <= '\176')
	{
	    *txt = c;
	    *(txt + 1) = '\0';
	}
	else if (c < 32)
	{
	    sprintf (txt, "<%s>", ccname[(int) c]);
	}
	else
	{
	    sprintf (txt, "<%02x>", c);
	}
	if (LOG)
	{
	    fputs (txt, LOG);
	}
	if (STDE)
	{
	    fputs (txt, stderr);
	}
    }
}

void logs (char *p)
{
    short n = 0;
    char c;

    if (LOG || STDE)
    {
	for (; p && *p; p++)
	{
	    if (achar)
	    {
		c = *p;
		if (c >= ' ' && c <= '\176')
		{
		    if (LOG)
		    {
			fputc (c, LOG);
		    }
		    if (STDE)
		    {
			fputc (c, stderr);
		    }
		    n++;
		}
	    }
	    else
	    {
		logc (*p);
		n++;
	    }
	}
	if (n)
	{
	    if (LOG)
	    {
		fputc ('\n', LOG);
		fflush (LOG);
	    }
	    if (STDE)
	    {
		fputc ('\n', stderr);
		fflush (stderr);
	    }
	}
    }
}

QFAXHEADER *SetHeader (QFAXHEADER * h, PT30PARAMS p, int hint, 
		       short pmax, int fd)
{
    memset (h, 0, QFAXHEADLEN);
    if (p)
    {
	h->title = QFAXTITLE;
	h->version = QFAXVERS;
	h->width = p->wd;
	h->resolution = p->vr;
	h->coding = p->df;
	h->page = p->ln;
	tsi[23] = 0;
	strcpy (h->tsi, tsi);
	h->stamp = faxtime;
	h->pagemax = pmax;
	h->hint = hint;
	h->lsb = my.lsb & 1;
	HeaderEndianise(h);
	if (PgH.head)
	{
	    MLIST *m, *lm;

	    h->offset = lseek (fd, 0, 1);
	    h->offset = LG(h->offset);
	    for (m = PgH.head; m;)
	    {
		OffrecEndianise((OFFREC *)m->udata);
		write (fd, m->udata, sizeof (OFFREC));
		lm = m;
		m = m->next;
		free (lm);
	    }
	    PgH.head = PgH.tail = NULL;
	}
    }
    return ++h;
}

QFAXHEADER *ReadHeader (FILE * f, short flag)
{
    static QFAXHEADER h;
    struct stat s;

    fstat (fileno (f), &s);

    fread (&h, 1, QFAXHEADLEN, f);
    HeaderEndianise(&h);
    if (h.title != QFAXTITLE || h.version < QFAXNEWFMT)
    {
	fputs ("Not v2.0 QFAX\n", stderr);
	exit (QX_ERR_5);
    }

    if (h.version <= QFAXFOFF)
    {
	h.offset = 0;
    }

    if (h.offset)
    {
	OFFREC o;

	fseek (f, h.offset, 0);
	while (fread (&o, 1, sizeof (o), f) == sizeof (o))
	{
	    MLIST *ml;

	    ml = malloc (sizeof (MLIST *) + sizeof (OFFREC));
	    OffrecEndianise(&o);
	    memcpy (ml->udata, &o, sizeof (o));
	    AppendList (&PgH, ml);
	}
	fseek (f, sizeof (h), 0);
	eofPos = h.offset;
    }
    else
    {
	eofPos = s.st_size;
    }

    if (flag)
    {
	int i;

	for (i = 0; i < MAXIOPT; i++)
	{
	    char *s;

	    if ((s = *(iopt + i)) && strncmp (s, fdcc[c20], 4) == 0)
	    {
		if (*qd.unlimited)
		{
		    h.page = 2;
		}
		my.res = h.resolution;
		my.code = h.coding;
		my.page = h.page;
		my.width = h.width;
		                                   
		sprintf (s, "%s=%d,%d,%d,%d,%d,%d,0,0",   
			 fdcc[c20],
			 h.resolution, 
			 (bopt != -1) ? bopt : qd.vmodem, 
			 my.width, my.page, my.code, (qd.ec[0] & 1));
		break;
	    }

	}
    }
    if (h.version > QFAXFLSB)
    {
	h.lsb &= 1;
    }
    else
    {
	h.lsb = 0;
    }
    return &h;
}

/*
 * Read a string from the modem.  The string is a CR-delimited modem command
 * if it begins with 'A', else it is a LF-delimited modem response.
 * Returns 0 if OK, +ve if ESCaped (msg, hit), -ve for error
 */

static int faxgets (char *s, int n, int t, int mflag)
{
    int c = 0;
    char *p;
    char termc = '\n';
    short tlm, tc;
    short first = 1;
    int q = 0;

    p = s;
    *p = 0;
    tlm = t * 10;
    tc = 0;
    n--;

    while (n)
    {
#ifdef QDOS
	if (mflag)
	{
	    if (domon == 2)
	    {
		if (popres == 0)
		{
		    q = 1;
		    break;
		}
	    }
	}
#endif
	c = SRead (p, 1, 1);
	if (c < 0)
	{
	    q = ERR_CTRL_C;
	    break;
	}
	else if (c == 0)
	{
#ifdef QDOS
	    if (Iserve && srvmsg == NULL)
	    {
		if ((ctx = SrvPoll ("QFAX", 0, &srvmsg, &srvlen)) >= 0)
		{
		    q = 2;
		    break;
		}
	    }
#endif
	    if (tc++ == tlm)
	    {
		q = ERR_TIMEOUT;
		break;
	    }
	}
	else
	{
	    if (first)
	    {
		if (tolower (*p) == 'a')
		{
		    termc = '\r';
		}
		first = 0;
	    }
	    n--;
	    if (*p++ == termc)
	    {
		break;
	    }
	}
    }
    *p = 0;
    return q;
}

/* Search for a match to the string s in a null-delimited table of prefix * * 
 * strings pointed to by p. Returns pointer to the table entry or NULL.  */

char *strtabmatch (char **p, char *s)
{
    while (*p && strncmp (*p, s, strlen (*p)))
	p++;
    return *p;
}

void ParseTSI (void)
{
    char *s, *r;
    isanon = 0;
    
    alias = 0;

    for (s = tsi, r = ttsi; *s; s++)
    {
	if (isalnum (*s))
	{
	    *r++ = tolower(*s);
	}
    }
    *r = '\0';
    if (*ttsi == '\0')
    {
	strcpy (ttsi, "unknown");
        isanon = 1;
    }
    else
    {
	if (qd.alias)
	{
	    ALIAS *pl;

	    for (pl = qd.alias; pl && !alias; pl = pl->next)
	    {
		if (strcmp (ttsi, pl->phone) == 0)
		{
		    strcpy (ttsi, pl->alias);
		    strlwr(ttsi);
		    alias = 1;
		}
	    }
	}
    }
}

/*
 * Send command to modem and check responses.  Writes the command s to the
 * modem if s is not null.  Notes the values of important responses (FPTS:
 * Page Transfer Status, and FHNG: HaNGup status).  Terminates when a
 * response is one of the prompts in promptstr[].  Returns the first 
 * character of the last response from the modem (e.g. 'O' for OK, 'C' for
 * CONNECT, etc.)  or EOF if no such response was received within t
 * seconds.
 */

#ifdef QDOS
char *stpcpy (char *d, const char *s)
{
    while (*d++ = *s++) ;
    return (d - 1);
}
#endif
static short lastin;
static char *cmd (char *s, int t)
{
    static char buf[CMDBUFSIZE];
    int p;
    char *q;

    if (s)
    {
	msg (0, "modem command [%s]", s ? s : "(none)");
	if (qd.delay)
	{
	    Delay (qd.delay);
	}
	if (qd.iwait == 0)
	{
	    SWrite ("AT", -1);
	    SWrite (s, -1);
	    SWrite ("\r", -1);
	}
	else
	{
	    q = stpcpy (buf, "AT");
	    q = stpcpy (q, s);
	    *q++ = CR;
	    *q = 0;
	    for (q = buf; *q; q++)
	    {
		Delay (qd.iwait);
		SWrite (q, 1);
	    }
	}
    }
    
    while ((p = faxgets (buf, CMDBUFSIZE, t, 0)) == 0)
    {
	logs (buf);
	if (strtabmatch (promptstr, buf))
	    break;
	ParseResponse (buf);
	lastin = *buf;
    }
    return (p == 0) ? buf : "";
}

/*
 * Reset the modem from an unknown state.
 * See if responds to command s.  If not, cancel possible fax send or
 * receive modes and drain any pending data or messages. If it still doesn't
 * work, try the default on-line escape sequence.  Returns 0 if OK or 4
 * if no response.
 */

int faxsync (char *s)
{
    int err = 0;
    int c;

    while ((c = *cmd (0, 0)))
	;

    {
	if ((c = *cmd (s, 2)) != OK && c != 'N')
	{
	    Delay (50);
	    SWrite ("+++", -1);
	    Delay (50);
	    if (*cmd (fk[c20], 2) != OK)
		err = msg (QX_ERR_4, "modem not responding");
	    else
		cmd (s, 2);
	}
    }
    return err;
}
#ifdef QDOS
void KillPopUp (void)
{
    mt_frjob (poppid, 0);
    poppid = 0;
    popres = -1;
}
#endif

/* signal handler: hang up and exit */


int end_session (char *s)
{
    int err = faxsync (s);

    return err;
}

void onsig (int sig)
{
#ifdef QDOS
    if (poppid)
    {
	KillPopUp ();
    }
#endif
    msg (0, "terminating on signal %d", sig);
    end_session ("H");
#if defined (__linux__) || defined (__NT__)
    TidyUp();
#endif
    exit (QX_ERR_4);
}


/* Initialize session.  Open fax device. Register signal handler. Set tty * * 
 * modes, modem options, fax mode and capabilities.  Then use specified * *
 * command (dial/answer) to start the session.  */

void ConnectMsg (void)
{
    char txt[160];

    sprintf (txt, "Connect at %s bps to %s", t30_br_string (&t30p), tsi);
    if (alias)
    {
	int n;

	n = strlen (txt);
	sprintf (txt + n, " (%s)", ttsi);
    }
    msg (QX_FORCE, txt);
    my.rate = t30p.br;
}

static short openTTY = 1;
static long pSer = 0;

int begin_session (int init, short fclass)
{
    int i, err = 0;
    char c;
    char *s;

    if (openTTY)
    {
#if defined (__linux__) 
	{
	    char *d;
	    d = (act == 'z') ? NULL : qd.serdev;
	    pSer = tty_open(d, (init) ? qd.baud : 0, init);
	}
#elif defined (__NT__)
	pSer = tty_open (qd.serdev, (init) ? qd.baud : 0, init);
#else
#ifdef QDOS
	if (qd.sbaud && (act == 's' || act == 'S' || act == 'z'))
	{
	    msg(0, "Using send baud %d (recv = %d)", qd.sbaud, qd.baud);
	    qd.baud = qd.sbaud;
	}
#endif
	pSer = tty_open (init);
#endif
	if(pSer >= 0)
	{
	    openTTY = 0;
	    signal (SIGABRT, onsig);
	    signal (SIGTERM, onsig);
	    signal (SIGINT, onsig);
	    if (init)
	    {
		err = faxsync (inicmd);
	    }
	}
	else exit(0);
    }

    if ((s = strstr (iopt[FX_FCLASS], "+FCLASS=")))
    {
	*(s + 8) = '0' + fclass;
    }

    c = tolower (act);
    if (c != 's' && c != 'z')
    {
	if ((s = strstr (iopt[FX_FAA], "+FAA=")))
	{
	    *(s + 5) = faa + '0';
	}
    }

    if (my.lsb & 1)
    {
	if ((s = strstr (iopt[FX_FAA], fbor[c20])))
	{
	    *(s + FBORLEN-c20) = (c20 == 0) ? '1' : '0';
	}
    }

    for (i = 0; !err && i < MAXIOPT; i++)
    {
	if (iopt[i])
	{
	    int j;

	    for (j = 0; j < 2; j++)
	    {
		c = *cmd (iopt[i], (j) ? 2 : 10);
		if (c != OK)
		{
		    err = msg (QX_ERR_3, "modem initialization command (%s) failed", iopt[i]);
		}
		else
		{
		    err = 0;
		    break;
		}
	    }
	}
    }
    hsc = -1;
    return err;
}


void CloseDownSerial (void)
{
    tty_close ();
    openTTY = 1;
}
void ReopenSerial (void)
{
    begin_session (1, qd.rclass);
}

#ifdef __linux__

void DataCmd(char *cmd)
{
    int status;
    char *ptr;
    char fcmd[128];
    pid_t cpid;
    short flg = 0;
    char **pa, *args[64];
    char *p1, *p2;

    sprintf(fcmd, cmd, qd.baud);
    
    if(*fcmd == '-')
    {
	ptr = fcmd+1;
	flg = 1;
    }
    else
    {
	ptr = fcmd;
    }

    if(*ptr == '-')
    {
	ptr++;
	flg |= 2;
    }

    pa = args;
    for(p1 = ptr; (p2 = strtok(p1, "\t ")); p1 = NULL)
    {
	*pa++ = p2;
    }
    *pa = NULL;

    switch((cpid = fork()))
    {
      case 0:
	if(flg & 1)
	{
	    dup2(pSer, 0);
	}
	if(flg & 2)
	{
	    dup2(pSer, 1);
	}
	execvp (args[0], args);
	_exit(0);
	break;
      case -1:
	_exit(EXIT_FAILURE);
	break;
      default:
	waitpid(cpid, &status, 0);
	break;
    }
}

int AsyncSystem (char *command)
{
#if 0
    pid_t pid;
    char **pa, *args[64];
    char *p1, *p2;
    
    pa = args;
    for(p1 = command; (p2 = strtok(p1, "\t ")); p1 = NULL)
    {
	*pa++ = p2;
    }
    *pa = NULL;

    if ((pid = fork ()) == 0)
    {
	execvp (args[0], args);
	_exit(0);
    }
    else if (pid < 0)
    {
         return -1;
    }
    else
    {
	return pid;
    }
#else
    char cmd[256];
    sprintf(cmd, "%s &", command);
    system(cmd);
    return 0;
#endif

}    

#else
#ifdef _DCC
void AsyncSystem (char *command)
{
    struct TagItem stags[] =
    {
	{SYS_Input, 0},
	{SYS_Output, 0},
	{SYS_Asynch, TRUE},
	{NP_Priority, -10},
	{NP_StackSize, 20000},
	{TAG_DONE}};

    System (command, stags);
}
#else
#ifdef QDOS
extern int SpawnCmd (char *, char *);
void AsyncSystem (char *command)
{
    char *p;
    char *opts = NULL;

    p = strchr (command, ' ');
    if (p)
    {
	*p = '\0';
	opts = p + 1;
    }
    SpawnCmd (command, opts);
}
#else
extern void AsyncSystem (char *command);
#endif
#endif
#endif

void PrintFaxFile (char *fname)
{
    FILE *fp;
#if !defined (__linux__) && !defined (__NT__)
    char plist[128], qnam[80];
#else
    char plist[512], qnam[256];
#endif
    char *pfnam,*p;

#ifdef _DCC
    pfnam = stpcpy (plist, qd.pcmd);
    *pfnam++ = ' ';
    strcpy (pfnam, qd.in_dir);
    AddPart (pfnam, fname, (sizeof (plist) - (pfnam - plist)));
    strcpy (qnam, qd.in_dir);
    AddPart (qnam, "QFAX-LastFax", sizeof (qnam));
#else
    pfnam = fname;
    p = stpcpy (plist, qd.pcmd);
    *p++ = ' ';
#if defined (__linux__) || defined (__NT__)
    p = stpcpy (p, qd.in_dir);
#endif
    stpcpy (p, fname);
    sprintf (qnam, "%sQFAX-LastFax", qd.in_dir);
#endif

    if ((fp = fopen (qnam, "w")))
    {
	strcpy (LastNam, pfnam);
#if defined (__linux__) || defined (__NT__)
	fputs(qd.in_dir, fp);
#endif
	fputs (pfnam, fp);
	fputc ('\n', fp);
	fclose (fp);
    }
    if (prt && *qd.pcmd)
    {
	msg (0, plist);
	AsyncSystem (plist);
    }
#ifndef QDOS
    if (*qd.arrivals)
    {
	struct fx
	{
	    char path[256];
	    time_t t;
	}
	*p, *q;
	int fd;

	if ((fd = open (qd.arrivals, O_RDWR|O_BINARY, 0)) > 0)
	{
	    struct stat s;

	    fstat (fd, &s);
	    if ((p = (struct fx *) malloc (s.st_size)))
	    {
		short nfx;

		read (fd, p, s.st_size);
		nfx = s.st_size / sizeof (struct fx);

		q = p + nfx - 1;
		strcpy (q->path, pfnam);
		q->t = faxtime;
		lseek (fd, 0, 0);
		write (fd, q, sizeof (struct fx));
		write (fd, p, s.st_size - sizeof (struct fx));

		free (p);
	    }
	    close (fd);
	}
    }
#else
    {
	char *sfnam = NULL;

	if (*qd.arrivals)
	{
	    sfnam = qd.arrivals;
	}
	AddFaxList (pfnam, TIME_UNIX_QL (faxtime), sfnam);
    }
#endif
}

/*
 * Send data for one page. Sets up serial port.  Reads characters from file
 * until EOF or error.  Output characters are buffered. Characters are
 * bit-reversed and DLE's are escaped.  Buffer is written when full or if 
 * done.  On each write() to the fax device checks for cancel character (CAN) 
 * from the remote end. When done, sends DLE-ETX, returns serial port
 * to command mode.
 */


int send_fax_data (FILE * f, long maxb, short border, int *nsiz)
{
    int done = 0, err = 0, can = 0, c;
    unsigned char q;
    unsigned char *op;
    unsigned char obuf[SNDBUFSIZE + 5];
    short urev = qd.sendrev;
    int nb = 0;
    
    if (border)
    {
	urev ^= 1;
    }
    op = obuf;

    InitRTC (1, 6);
    
    while (!done)
    {
	if (maxb-- <= 0)
	{
	    done = 1;
	}
	else if ((c = fgetc (f)) == EOF)
	{
	    done = 1;
	    if (ferror (f))
		err = msg (QX_ERR_2, "fax file read: %s", strerror (errno));
	}

	if (!done)
	{
	    int xxx;
	    
	    if (urev)
	    {
		c = reversebits[c];
	    }

	    if ((xxx = CheckRTC (c)))
	    {
		msg(0, "Sending RTC %d", xxx);
		done = 1;
	    }

	    *op++ = c;
	    if (c == DLE)
		*op++ = DLE;
	}

	if (op >= (obuf + SNDBUFSIZE) || done)
	{
	    while (SRead ((char *) &q, 1, 0) == 1)
	    {
		if (q == CAN)
		{
		    done = can = 1;
		    err = msg (QX_ERR_3, "transmission CANcelled");
		}
		else
		{
		    if (q != XON && q != XOFF)
		    {
			msg (0, "[%02x] received while sending data", q);
		    }
		}
	    }
	    if (done && (c20 == 0))
	    {
		*op++ = DLE;
		*op++ = ETX;
	    }

	    if (can)
		*op++ = CAN;
	    {
		int nc;
		nc = (op - obuf);
		nb += nc;
		SWrite ((char *) obuf, nc);
	    }
	    op = obuf;
	}
    }
    *nsiz = nb;
    return err;
}

void CheckMode (QFAXHEADER * h, long fpage, short *diff)
{
    if (PgH.head)
    {
	MLIST *ml;
	short ucode = h->coding;
	short ures = h->resolution;
	char upage = h->page;
	char uwidth = h->width;

	for (ml = PgH.head; ml; ml = ml->next)
	{
	    if (abs (fpage - ((OFFREC *) ml->udata)->off) < 4)
	    {
		ucode = ((OFFREC *) ml->udata)->code;
		ures = ((OFFREC *) ml->udata)->res;
		if (!qd.unlimited)
		{
		    upage = ((OFFREC *) ml->udata)->page;
		}
		uwidth = ((OFFREC *) ml->udata)->width;
	    }
	    break;
	}

	if (ucode != my.code || ures != my.res ||
	    upage != my.page || uwidth != my.width)
	{
	    *diff = 1;
	    my.code = ucode;
	    my.res = ures;
	    my.page = upage;
	    my.width = uwidth;
	}
    }
}
/*
 * Send a page. Open the image file for binary reading.  Give the data
 * transmit command. Transmission begins after DC2 (or XON) is received.
 * Sends the data, sends appropriate post-page message, and closes the image
 * file.
 */

int send_page (FILE * f, int pg, QFAXHEADER * h, long fpage, 
	       int *tbytes, int *ttim)
{
    int err = 0;
    char c;
    char s[64];
    short last = (pg == h->pagemax - 1);
    long maxb;
    short diff = 0;
    static short force = 0;
    short lc;
    time_t t1;
    int nsiz;
    
    maxb = eofPos - fpage;

    if (pg == 0)
    {
	if (*qd.unlimited)
	{
	    h->page = 2;
	}

	if (my.res != h->resolution || my.code != h->coding ||
	    my.page != h->page || my.width != h->width)
	{
	    my.res = h->resolution;
	    my.code = h->coding;
	    my.page = h->page;
	    my.width = h->width;
	    diff = 1;
	}
    }

    CheckMode (h, fpage, &diff);

    if (diff || force)
    {
	sprintf (s, "%s=%d,%d,%d,%d,%d,%d,0,0", fdcc[c20],
		 my.res, my.rate, my.width, my.page, my.code, (qd.ec[0] & 1));
	err = (*cmd (s, 10) != OK);
	force = 0;
    }

    if (!err && *cmd ("+FDT", 40) != CONNECT)
    {
	err = msg (QX_ERR_3, "send (+FDT) command failed");
    }
    else if (pg == 0)
    {
	ConnectMsg ();
    }


    if (c20 == 0 && !err && qd.wait2send)
    {
	while (!err)
	{
	    msg (0, "Waiting for start");
	    if (SRead (&c, 1, 200) == 0)
	    {
		err = msg (QX_ERR_3, "timed out waiting to send data");
	    }
	    else if (c == XON || c == DC2)
	    {
		break;
	    }
	    else
	    {
		msg (0, "[%02x] received while waiting to send data", c);
	    }
	}
    }
    if (!err)
    {
	t1 = time(NULL);
	msg (0, "Sending page %d ...", pg + 1);
	err = send_fax_data (f, maxb, h->lsb, &nsiz);
	if (!last)
	{
	    lc = 0;
	    CheckMode (h, ftell (f), &lc);
	    force = lc;
	}
	else
	{
	    if (poll)
	    {
		lc = 1;
	    }
	    else
	    {
		lc = 2;
	    }
	}
	if(c20)
	{
	    SWrite(c2ppm[lc], 2);
	    msg(0, "sending lc = %d, [%x,%c]", lc, 
		c2ppm[lc][0], c2ppm[lc][1]);
	}
	
    }
    if (!err)
    {
	if (*cmd (0, 40) != OK)
	{
	    err = msg (QX_ERR_3, "no response after sending data");
	}
	else
	{
	    char fet[16];

	    (*tbytes) += nsiz;
	    (*ttim) += time(NULL) - t1;
	    if(c20 == 0)
	    {
		sprintf (fet, "+FET=%d", lc);
		if (*cmd (fet, 30) != OK)
		{
		    err = msg (QX_ERR_3, "no post-page-response");
		}
	    }
	}
    }
    return err;
}

void DoFaxLog (char *fact, short page, short err)
{
    if (*qd.info_file)
    {
	FILE *fp;
	time_t t;

	t = faxtime;
	if (t == 0)
	{
	    t = time (NULL);
	}
	if ((fp = fopen (qd.info_file, "a")) != NULL)
	{
	    fprintf (fp, "%.24s %s %d page(s), %s, result %s\n",
		     ctime (&t), fact,
		     (page > 0) ? page : 0,
		     tsi,
		     (err) ? "FAIL" : "OK");

	    fprintf (fp, "Connect %d secs, br: %s vr: %s wd: %s ln: %s df: %s",
		     online,
		     t30_br_string (&t30p),
		     t30_vr_string (&t30p),
		     t30_wd_string (&t30p),
		     t30_ln_string (&t30p),
		     t30_df_string (&t30p));

	    if (x30p.df == 1 && t30p.df == 0)
	    {
		fputs (" ***  supports 2D\n", fp);
	    }
	    else
	    {
		fputc ('\n', fp);
	    }
	    fclose (fp);
	}
    }
}

void FindCSI (char *csi)
{
    alias = 0;
    if (qd.alias)
    {
	ALIAS *pl;

	for (pl = qd.alias; pl && !alias; pl = pl->next)
	{
	    if (strcasecmp (csi, pl->abbrev) == 0 ||
		strcasecmp (csi, pl->alias) == 0)
	    {
		strcpy (ttsi, pl->alias);
		strlwr(ttsi);
		if (pl->flag & 0x8000)
		{
		    strcpy (csi, pl->dcode);
		}
		else
		{
		    if (*pl->dcode)
		    {
			if (*pl->dcode == '-')
			{
			    strcpy (csi, pl->dcode + 1);
			    break;
			}
			strcpy (csi, pl->dcode);
		    }
		    else
		    {
			*csi = '\0';
		    }
		    strcat (csi, pl->phone);
		}
		alias = 1;
	    }
	}
    }
}


/*
 * Dial a number and send a fax.  Sends files fname[0...n-1].
 */

int DialNumber (char *tel)
{
    int err = 0;
    char c;
    char telno[128];
    char s[128];

    strcpy (telno, tel);
    if (*telno == '!' || *telno == '*' || *telno == '_')
    {
	*telno = 0;
    }
    else
    {
	if (isalpha (*telno))
	{
	    FindCSI (telno);
	    if (!alias)
	    {
		err = QX_ERR_5;
	    }
	}
    }

    err = begin_session (act != 'z', qd.sclass);

    if (!err)
    {
	if (*qd.poll)
	{
	    if (ipoll)
	    {
		short perr = 0;

		perr = (*cmd (fspl1[c20], 5) != OK);
		if (!perr)
		{
		    sprintf (s, "+FCIG=%s", qd.LocalId);
		    err = (*cmd (s, 5) != OK);
		}
	    }
	    else
	    {
		cmd (fspl0[c20], 4);
	    }
	}
    }

    if (!err)
    {
	char *osb;

	if ((osb = strchr (telno, '[')))
	{
	    if (*qd.indial)
	    {
		char *sv;

		if ((sv = strdup (osb + 1)))
		{
		    strcpy (osb, qd.indial);
		    strcat (telno, sv);
		    free (sv);
		}
	    }
	    else
	    {
		*osb = ' ';
	    }
	}
	sprintf (s, "%s%.127s", dialcmd, telno);
	c = *cmd (s, 120);

	if (c == BUSY)
	{
	    err = msg (QX_ERR_1, "number is busy");
	}
	else if (c == OK && hsc < 0)
	{
	    msg (0, "connection established");
	    if (my.res > x30p.vr || my.code > x30p.df)
	    {
		err = msg (QX_ERR_4, "Respondent can't cope !!");
		cmd (fk[c20], 5);
		cmd ("H", 5);
	    }
	}
	else
	{
	    err = msg (QX_ERR_3, "can't establish session");
	}
    }

    return err;
}

extern int DoFaxReceive (char *, char *);

int FaxSend (char *tel, char *fname)
{
    int err = 0, page, try;
    FILE *f;
    QFAXHEADER *h;
    int fpage;
    int tbytes = 0, ttim = 0;

    online = 0;

    h = NULL;
    f = NULL;

    if (fname)
    {
	if ((f = fopen (fname, "rb")) == NULL)
	{
	    err = msg (QX_ERR_5, "fax file open %s %s", strerror (errno), fname);
	    if (tel == NULL)
	    {
		cmd (fk[c20], 5);
	    }
	}
	else
	{
	    msg (0, "opened fax file %s", fname);
	    h = ReadHeader (f, (tel) ? 1 : 0);
	}
    }
    else
    {
	if (ipoll == 0)
	{
	    err = msg (QX_ERR_5, "no file to send");
	}
    }

    if (!err && tel)
    {
	err = DialNumber (tel);
	if (h)
	{
	    qmode = "Sending";
	    if (!err)
	    {
		msg (0, "sending %d pages to %s", h->pagemax, tel);
	    }
	}
	else
	{
	    qmode = "Poll request";
	    if (!err)
	    {
		msg (0, "Polling %s", tel);
	    }
	}
    }

    if (f && !err)
    {
	if (tel == NULL)
	{
	    msg (0, "sending %d pages from poll fax %s", h->pagemax, pollfile);
	    qmode = "Polled Send";
	}
	online = faxtime = time (NULL);

	for (page = 0; !err && page < h->pagemax; page++)
	{
	    for (try = 0; !err && hsc < 0; try++)
	    {
		good = 1;
		fpage = ftell (f);
		err = send_page (f, page, h, fpage, &tbytes, &ttim);

		if (good)
		{
		    break;
		}
		else
		{
		    if (try < qd.ppr)
		    {
			msg (1, "Retry for ppr = %d", ppr);
			fseek (f, fpage, SEEK_SET);
		    }
		    else
		    {
			break;
		    }
		}
	    }

	    if (!err && !good)
		err = msg (QX_ERR_3, "too many page send retries");
	    if (hsc > 0)
		err = msg (QX_ERR_3, "abnormal call termination (code %d)", hsc);
	}

	if (online && !poll)
	{
	    online = time (NULL) - online;
	}
	DoFaxLog (qmode, page, err);
	if (x30p.df == 1 && t30p.df == 0)
	{
	    msg (0, "Note: respondent supports 2D");
	}

	if (PgH.head)
	{
	    MLIST *m, *lm;

	    for (m = PgH.head; m;)
	    {
		lm = m;
		m = m->next;
		free (lm);
	    }
	    PgH.head = PgH.tail = NULL;
	}
    }

    if(f)
    {
	if (fclose (f) != 0)
	{
	    msg (0, "fax file close: %s", strerror (errno));
	}
	else
	{
	    msg (0, "closed fax file");
	}
    }

    if(ttim)
    {
	msg(0, "Sent %d bytes in %d secs (tx rate = %d cps)", 
	    tbytes, ttim, tbytes/ttim);
    }
    
    if (poll)
    {
	char pname[128];

	*pname = '\0';
	qmode = "Receive Poll";
	if (!online)
	{
	    online = faxtime = time (NULL);
	}
	ParseTSI ();		/* 10/12/94 JH */

	err = DoFaxReceive (NULL, pname);
	if (err == 0 && *pname)
	{
	    PrintFaxFile (pname);
	}
    }

    if (ptsi == 0 && abs (err) < QX_ERR_4)
    {
	end_session ("H");
    }

    poll = ipoll = 0;
    return err;
}

#define MAXNULLS 5

#ifdef QDOS
# define MAXRBUF 8*1024
#else
# define MAXRBUF 64*1024
#endif

#define OFFSET_MAXPAGE ((long)&((QFAXHEADER *)0)->pagemax)

/*
 * Receive a fax.  Open file for binary write.  Send fax data receive
 * command.  If response is OK, then no more pages are coming.  If it's 
 * CONNECT receive the data for one page. Returns 0 if a page was received,
 * 1 if no more pages, 2 or more for errors.
 */


static short conf;

int receive_fax (char *fname, short *page)
{
    int err = 0;
    int fd;
    char *buf;
    int rb;
    int done = 0, dle = 0;
    unsigned char c;
    int ochar;
    short nulls = 0;
    unsigned char p1, p2, p3;
    int rtc;
    int maxrtc = 0;
    time_t t1,t2;
    int nsiz, tbytes = 0, ttim = 0;
    
    p1 = p2 = p3 = 0;

    *page = 0;
    fd = open (fname, O_WRONLY | O_BINARY | O_CREAT, S_IRUSR|S_IRGRP|S_IROTH|S_IWUSR);
    if (fd < 0)
    {
	err = msg (QX_ERR_2, "fax file open %s\n%s:", strerror (errno), fname);
    }
    else
    {
	msg (0, "opened fax file %s", fname);
	for (rb = MAXRBUF; rb > 512; rb = rb >> 1)
	{
	    if ((buf = malloc (rb)))
	    {
		break;
	    }
	}
    
	ochar = QFAXHEADLEN;
	do
	{
	    short timo;
	    rtc = 0;
	    timo = qd.hangup[0] ? 120 : 60;
	    if ((c = *cmd ("+FDR", timo)) != CONNECT)
	    {
		if (c == OK)
		{
		    err = QX_DONE;
		}
		else
		{
		    if(qd.hangup[0] && hsc == 0)
		    {
			msg (0, "HANGUPOK set - no OK after (+FDR)");
			err = QX_DONE;
		    }
		    else
			err = msg (QX_ERR_3, "receive (+FDR) command failed");
		}
	    }
	    
	    if (!err)
	    {
		if (*page == 0)
		{
		    ini30p = t30p;
		    l30p = t30p;
		}
		else
		{
		    if (l30p.vr != t30p.vr ||
			l30p.ln != t30p.ln ||
			l30p.df != t30p.df
			|| l30p.wd != t30p.wd)
		    {
			MLIST *ml;

			ml = malloc (sizeof (MLIST *) + sizeof (OFFREC));
			((OFFREC *) ml->udata)->off = lseek (fd, 0, 1);
			((OFFREC *) ml->udata)->pageno = *page + 1;
			((OFFREC *) ml->udata)->page = t30p.ln;
			((OFFREC *) ml->udata)->page = t30p.wd;
			((OFFREC *) ml->udata)->code = t30p.df;
			((OFFREC *) ml->udata)->res = t30p.vr;
			AppendList (&PgH, ml);
		    }
		    l30p = t30p;
		}
		if (qd.startchar)
		{
		    msg (0, "Send start char <%x>", qd.startchar);
		    SWrite (&qd.startchar, 1);
		}
		rtc = 0;
		t1 = time(NULL);
		nsiz = 0;
		
		for (done = 0; done == 0;)
		{
		    int nc;
		    if ((nc = SRead ((char *) &c, 1, 250)) == 1)
		    {
			nsiz++;
			if (c)
			{
			    nulls = 0;
			    if (p1 == 0)
			    {
				p1 = c;
				if (my.lsb == 0)
				{
				    if (c & 0x80)
				    {
					my.lsb = 1;
				    }
				    else if (c & 0x01)
				    {
					my.lsb = 0;
				    }
				    else if (c & 0x10)
				    {
					my.lsb = 0;
					conf = 0;
				    }
				    else if (c & 0x08)
				    {
					my.lsb = 1;
					conf = 0;
				    }
				    else
				    {
					msg (0, "Unable to detect order (%lx)", c);
				    }
                                    my.lsb &= 1;
				    msg (0, "Bit order setting (%csb)", 'm' - my.lsb);
				}
                		else
				{
                                    my.lsb &= 1;
				    msg (0, "Bit order override (%csb)", 'm' - my.lsb);
				}
				InitRTC (my.lsb, 6);
			    }
			}
			else if (++nulls > MAXNULLS)
			{
			    continue;
			}

			if (dle)
			{
			    if (c != DLE)
			    {
				dle = 0;
				if (c == ETX)
				{
				    msg (0, "End of page detected");
				    done = 1;
				}
				else
				{
				    msg (0, "[%02x] received after DLE", c);
				}
				continue;
			    }
			}

			if (c == DLE)
			{
			    if (dle)
			    {
				dle = 0;
			    }
			    else
			    {
				dle = 1;
				continue;
			    }
			}
			if (CheckRTC (c))
			{
			    rtc = GetFaxRTC (NULL);
			}
			*(buf + ochar) = c;
			ochar++;

			if (ochar == rb)
			{
			    write (fd, (char *) buf, rb);
			    ochar = 0;
			}
		    }
		    else
		    {
#ifdef _DCC
			if (((struct IOExtSer *) pSer)->IOSer.io_Error == 0)
			{
			    done = 1;
			    err = msg (QX_ERR_3, "fax device read timeout");
			}
			else
			{
			    msg (QX_ERR_3, "serial error %ld", ((struct IOExtSer *) pSer)->IOSer.io_Error);
			}
#else
			done = 1;
			err = msg (QX_ERR_3, "fax device read timeout");
#endif
		    }
		}

		if (ochar > 0)
		{
		    write (fd, (char *) buf, ochar);
		    ochar = 0;
		}
		(*page)++;


		if (rtc == 0)
		{
		    short lz;
		    rtc = GetFaxRTC (&lz);
		    msg (0, "Adding RTC to fax page (%d %d)", rtc, lz);
		    write (fd, (char *) RTCDAT, RTCLEN);
		}

		c = *cmd (0, 20);

		t2 = time(NULL) - t1;
		tbytes += nsiz;
		ttim += t2;

		if (fline && fline > rtc)
		{
		    rtc = fline;
		}
		if (*page && (t30p.vr != ini30p.vr))
		{
		    if (ini30p.vr)
		    {
			rtc = rtc << 1;
		    }
		    else
		    {
			rtc = (rtc + 1) >> 1;
		    }
		}
		if (rtc > maxrtc)
		{
		    maxrtc = rtc;
		}
		if (c != OK)
		{
		    err = msg (QX_ERR_3, "no response after receiving data");
		}
		else
		{
		    msg (0, "page received, %d lines", rtc);
		}
		InitRTC (my.lsb, 6);
	    }
	} while (!err);

	if(ttim)
	{
	    msg(0, "Received %d bytes in %d secs, (rx rate %d cps)",
		tbytes, ttim, tbytes/ttim);
	}
	SetHeader ((QFAXHEADER *) buf, &ini30p, maxrtc, (*page), fd);
	lseek (fd, 0, 0);
	write (fd, (QFAXHEADER *) buf, QFAXHEADLEN);
	free (buf);
	close (fd);
    }
    return err;
}

/*
 * Answer phone and receive a fax.  Makes one file per page.  The file name
 * is built by appending a 3-digit number to the given string.
 */

#ifdef QDOS
#define SEPCHAR '_'
#define SEQDAT "seq_dat"
#else
#define SEPCHAR '.'
#define SEQDAT "seq.dat"
#endif


int rwait (char *rings)
{
    int err = 0, i = 0, n;
    char buf[CMDBUFSIZE], *p;
    int fr = 0;

#ifdef QDOS
    int dn = 0;

    if (domon && poppid)
    {
	KillPopUp ();
	domon = 1;
    }

#endif
    n = strtoul (rings, &p, 10);

    while (!err && i < n)
    {
	if ((fr = faxgets (buf, CMDBUFSIZE, RINGPERIOD, 1)) == 0)
	{
	    logs (buf);
	    if (strncmp (buf, RINGSTR, RINGSTRLEN) == 0)
	    {
#ifdef QDOS
		if (Iserve && i == 0)
		{
		    SrvBro ("QFAX", "Incoming Call", -1);
		}
		if(qd.edr && atoi(buf+RINGSTRLEN) == qd.edr)
		{
		    err = i = 0;
		    break;
		}
		if (domon && i == 0)
		{
		    if (poppid == 0)
		    {
			domon = 2;
			popres = -1;
			poppid = jcall (answer, "Take This Call ?", 1, 2048, 2, &poppid, &popres);
			if (poppid <= 0)
			{
			    domon = 0;
			}
		    }
		    dn = 0;
		}
#else
		if(qd.edr && atoi(buf+RINGSTRLEN) == qd.edr)
		{
		    err = i = 0;
		    break;
		}
#endif
		i++;
	    }
	}
	else
	{
#ifdef QDOS
/*---------------------------------------------------*/
	    if (domon && popres == 0)
	    {
		i = n;
		popres = -1;
	    }
	    else
	    {
		i = 0;
		if (poppid)
		{
		    dn++;
		    if (dn > 7)
		    {
			KillPopUp ();
			dn = 0;
		    }
		}
		else
		{
		    dn = 0;
		}
	    }

	    if (fr == 2)
	    {
		enum msg_states rcmd;

		rcmd = ActOnPipeMsg (ctx, &srvmsg, srvlen);

		if (rcmd == P_ANSWER)
		{
		    err = QX_OK;
		    i = n;
		}
		else if (rcmd == P_SEND)
		{
		    err = QX_ERR_SEND;
		}
		else if (rcmd == P_POLL)
		{
		    ipoll = 1;
		    err = QX_ERR_SEND;
		}
		else if (rcmd == P_GRAB)
		{
		    err = QX_ERR_GRAB;
		}
		else if (rcmd == P_BYE || rcmd == P_QUIT)
		{
		    exit (0);
		}
		else
		{
		    err = i = 0;
		}
	    }
/*--------------------------------------------*/
#else
	    switch (fr)
	    {
		case ERR_CTRL_C:
		    i = n;
		    err = QX_ERR_255;
		    break;
		case ERR_TIMEOUT:
		default:
		    i = 0;
		    break;
	    }
#endif
	}
    }
#ifdef QDOS
    if (poppid)
    {
	KillPopUp ();
    }
#endif
    return err;
}

int movedir (char *dirnam)
{
    int err = 0;

    if (chdir (dirnam) == -1)
    {
#ifdef _DCC
	mkdir (dirnam);
#else
	mkdir (dirnam, 0700);
#endif
	if (chdir (dirnam) == -1)
	{
	    err = msg (QX_ERR_2, "Error on directory %s", dirnam);
	}
    }
    return err;
}

int RecvDir (char *template, char *there, short thsize)
{
    int fd;
    int err = 0;
    unsigned short seq;

    getcwd (there, thsize);
    err = movedir (qd.in_dir);

    movedir (ttsi);

    fd = open (SEQDAT, O_RDWR | O_BINARY|O_CREAT, S_IRUSR|S_IRGRP|S_IROTH|S_IWUSR);
    if (read (fd, &seq, 2) != 2)
    {
	seq = 0;
    }
        // for hysterical reasons, this is also big endian. Anyone smell a
        // MC680x0 conspiracy here ?
    
    seq = SH(seq);
    sprintf (template, "%05d", seq);
    seq++;
    lseek (fd, 0, 0);
    seq = SH(seq);
    write (fd, &seq, 2);
    close (fd);

    return err;
}
int begin_grab (void)
{
    short err = 0;

    msg(-1, "Using zero");
#ifndef _DCC
#ifdef QDOS
    if (act == 'G')
    {
	tty_open(0);
    }
    else
    {
	tty_open(1);
    }
#else
/* Linux, open, don't change it */
    {
	tty_open (NULL, 0, 0);
    }
#endif
#else /* DCC */
    tty_open (0);
#endif
    signal (SIGABRT, onsig);
    signal (SIGTERM, onsig);
    signal (SIGINT, onsig);

    hsc = -1;
    if (*cmd (NULL, qd.gdelay) != 'O')
    {
	cmd (fk[c20], 5);
	err = msg (QX_ERR_3, "Grab fails");
    }
    else
    {
	msg (0, "establish GRAB session");
    }
    return err;
}

int DoSendFax (char *cmd)
{
    int err = QX_ERR_2;
    char *phno;
    char *s;

    qmode = "Sending";
    if ((s = strchr (cmd, '\n')))
    {
	*s = '\0';
    }
    phno = strtok (cmd + 5, " \t");
    s = strtok (NULL, " \t");
    if ((s && *s) || ipoll)
    {
	force = 0;
	err = FaxSend (phno, s);
	GetQFax ();
    }
    else
    {
	err = QX_ERR_2;
    }
    return err;
}
void end_data (void)
{
    char c;

    Delay (100);
    while (SRead (&c, 1, 5) == 1)
	;			/* !! */
    Delay (50);
    SWrite ("+++", 3);
    Delay (50);
    cmd ("H", 5);
}
int DoFaxReceive (char *base, char *fname)
{
    int err = 0;
    char template[80];
    char there[64];
    short page = 0;
    short cd;
    short via = 0;
    
    if (!err && base == NULL)
    {
	if (*qd.via == '\0')
	{
	    err = RecvDir ((base = template), there, 64);
	    cd = 1;
	}
	else
	{
	    sprintf ((base = template), "%s%d", qd.via, (int) faxtime);
	    via = 1;
	    cd = 0;
	}
    }
    else
    {
	cd = 0;
    }

    if (!err)
    {
	sprintf (fname, "%s%s", base, qd.suffix);
	msg (0, "receiving fax to %s", fname);
	err = receive_fax (fname, &page);
    }

    if (online)
    {
	online = time (NULL) - online;
    }

    if (err == QX_DONE)
	err = 0;

    if (via)
    {
	char newbas[80];
	char newfil[64];
	char *q;
	int j, f1, f2;

	if (!err)
	{
	    err = RecvDir (newbas, there, 64);
	    if (!err)
	    {
		cd = 1;
		if ((q = malloc (1024)))
		{
		    sprintf (newfil, "%s%s", newbas, qd.suffix);
#ifdef QDOS
		    if (strncmp (qd.in_dir, "nul_", 4))
#endif
		    {
			f1 = open (fname, O_RDONLY|O_BINARY);
			f2 = open (newfil, O_WRONLY | O_BINARY | O_CREAT, S_IRUSR|S_IRGRP|S_IROTH|S_IWUSR);
			while ((j = read (f1, q, 1024)) > 0)
			{
			    write (f2, q, j);
			}
			close (f1);
			close (f2);
		    }
		    msg (0, "Moved %s to %s", fname, newfil);
		    remove (fname);
		    free (q);
		}
	    }
	}
	base = newbas;
    }

    if (hsc > 0)
	err = msg (QX_ERR_3, "abnormal call termination (code %d)", hsc);

    if (cd)
    {
	chdir (there);
    }
#ifndef QDOS
    sprintf (fname, "%s/%s%s", ttsi, base, qd.suffix);
#else
    sprintf (fname, "%s%s_%s%s", qd.in_dir, ttsi, base, qd.suffix);
#endif
    DoFaxLog (qmode, page, err);
    return err;
}

int receive (char *base, char *rings)
{
    int err = 0;
    char fname[FILENAME_MAX + 1];

    online = 0;
    ptsi = 0;
    force = 0;

    qmode = "Receive";
#ifdef _DCC
    SetTaskPri (FindTask (0), 2);
#endif
    if (rings && *rings == 0)
    {
	err = begin_grab ();
    }
    else
    {
	if (qd.faa)
	{
	    faa = qd.faa;
	}
	err = begin_session (tolower (act) != 'z', qd.rclass);

	if (*qd.poll && *pollfile)
	{
	    err = (*cmd (flpl1[c20], 5) != OK);
	}
	if (!err && rings)
	{
	    err = rwait (rings);
	}
#ifdef QDOS
	if (err == QX_ERR_SEND)
	{
	    err = DoSendFax (srvmsg);
	    return err;
	}
	if (err == QX_ERR_GRAB)
	{
	    err = 0;
	}
	else
#endif
	{
	    if (!err)
	    {
		char *p = cmd ("A", 120);

		if (*p == 'C')
		{
		    char *p1;

		    if ((p1 = strstr (p, "CONNECT ")))
		    {
			if (strncasecmp (p1 + 8, "FAX", 3) == 0)
			{
			    p = cmd (NULL, 60);
			}
			else
			{
			    err = 0;
			    msg (0, "Attempting data connect");
			    qmode = "Data";
			    if (*qd.data)
			    {
				if(*qd.data == '@')
				{
				    FILE *fp;
				    msg(-1, qd.data);
				    if((fp = lopen(qd.data+1, "r")))
				    {
					char buf[256];
					while(fgets(buf, 250, fp))
					{
					    char *p;
					    if((p = strchr(buf, '\n')))
					    {
						*p++ = '\r';
						*p++ = '\n';
						*p = 0;
						SWrite(buf, -1);
					    }
					}
					fclose(fp);
				    }
				    else
				    {
					msg(-1,"Fails on %s", qd.data+1);
				    }
				}
				else
				{
				    
#ifdef _DCC
				    online = time (NULL);
				    SystemTags (qd.data, TAG_DONE);
#elif defined (QDOS)
				    int qchan;

				    online = time (NULL);
				    if (*qd.nul)
				    {
					qchan = OpenAppend (qd.nul, -1);
				    }
				    else
				    {
					qchan = fgetchid (STDE);
				    }
				    qsystem (pSer, qd.data, qchan);
				    if (*qd.nul && qchan)
				    {
					io_close (qchan);
				    }
#else
#ifndef __NT__
				    DataCmd (qd.data);
#else
  				    SWrite ("\r\nData is not available at this time\r\n", -1);
#endif
#endif
				}
				online = time (NULL) - online;
			    }
			    else
			    {
				SWrite ("\r\nData is not available at this time\r\n", -1);
			    }
			    end_data ();
			    return 0;
			}
		    }
		}
		if (*p != 'O')
		{
		    err = msg (QX_ERR_3, "modem answer failed");
		}
	    }
	}
    }

    if (!err)
    {
	int rerr = -1;

	online = faxtime = time (NULL);
	ParseTSI ();
        ConnectMsg ();
        if(!isanon || anon)
        {
            if (ptsi == 0)
            {
                *fname = '\0';
                rerr = err = DoFaxReceive (base, fname);
                force = 1;
            }
            if (ptsi == 1)
            {
                if (*pollfile)
                {
                    err = FaxSend (NULL, pollfile);
                }
            }
            if (rerr == 0 && *fname)
            {
                PrintFaxFile (fname);
            }
        }
        else
        {
            err = msg (QX_ERR_6, "No TSI on received fax, bailing out");
        }
    }

    if (abs (err) < 128)
    {
	end_session ("H");
    }

    return err;
}


char *CheckOpt (char *p)
{
    if (strncmp (p, "AT", 2) == 0)
    {
	p += 2;
    }
    return p;
}

void AddUserAT (char **pUserAT, char *p)
{
    p = CheckOpt (p);

    if (*pUserAT == NULL)
    {
	*pUserAT = malloc (512);
	**pUserAT = 0;
    }
    if (strstr (*pUserAT, p) == NULL)
    {
	if (**pUserAT != '\0')
	{
	    strcat (*pUserAT, " ");
	}
	strcat (*pUserAT, p);
    }
}

void fixupfdcc (void)
{
    static char mp1[80];    
    if (*qd.LocalId && *qd.LocalId != '_')
    {
	sprintf (mp1, "%s=%s", flid[c20], qd.LocalId);
	iopt[FX_FLID] = mp1;
    }
}

void fixupflid (void)
{
    static char mp2[80];    
    sprintf (mp2, "%s=%d,%d,%d,%d,%d,%d,0,0",
	     fdcc[c20], qd.res, my.rate,
	     qd.width, qd.page, qd.code, (qd.ec[0] & 1));
    iopt[FX_FDCC] = mp2;
}

void GetQFax (void)
{

    int i;

    FaxParams (&qd);
    if (*qd.data)
    {
	faa = 1;
    }

    if(*qd.twodot)
    {
	c20 = 1;
    }

    fixupfdcc ();

    if (qd.rmodem != -1)
    {
	my.rate = qd.rmodem;
    }
    else
    {
	my.rate = qd.vmodem;
    }

    my.res = qd.res;
    my.code = qd.code;
    my.width = qd.width;
    my.page = qd.page;
    my.lsb = qd.elsb[0];
    
    fixupflid ();

    if (*qd.init)
    {
	AddUserAT (iopt + FX_USERAT, qd.init);
    }

    if (*qd.preinit)
    {
	AddUserAT (iopt + FX_PREINI, qd.preinit);
    }

    dialcmd = CheckOpt (qd.dialcmd);
    inicmd = CheckOpt (qd.inicmd);

    for (i = 0; i < 4; i++)
    {
	if (*qd.ini[i] == -1)
	{
	    iopt[i + FX_FCLASS] = NULL;
	}
	else if (*qd.ini[i])
	{
	    iopt[i + FX_FCLASS] = CheckOpt (qd.ini[i]);
	}
    }
}

#ifdef __linux__
static char *lockfile = NULL;
static pid_t lockpid;
extern char *lock_dev(char *, pid_t *);
extern void unlock_dev(char *, pid_t);
#endif

#ifdef QDOS
static char xbaud;
static short brate[] =
{19200, 9600, 4800, 2400, 1200, 600, 300, 75};
/*   0    1    2    3    4   5   6  7 */
extern char *_sys_var;
#endif

void TidyUp (void)
{
    static short x = 1;
    if(x)
    {
	x = 0;
	FreeAlias (&qd);
#if defined(__linux__)
	if(pSer)
	{
	    tty_close();
	}

	if(lockfile)
	{
	    unlock_dev(lockfile, lockpid);
	}
	
#elif defined (QDOS)
	if (act != 'G' && tolower(act) != 'z')
	{
	    tty_close ();
	    if (qd.ebaud != -1)
	    {
		mt_baud (qd.ebaud);
	    }
	    else
	    {
		mt_baud (brate[xbaud]);
	    }
	    if ((qd.sflag & dact) && *qd.StopScript)
	    {
		RunScript (qd.StopScript, -1, ExitParams);
	    }
	}
	if (Iserve)
	{
	    ServerFree ();
	}
#endif
    }
}
#define QLERR_FLAG 0x80000000

int main (int ac, char **av)
{
    int i, err = 0, c;
    int loop = 0;
    char *base = NULL;
    char *rings = NULL;
#if defined(__linux__) || defined(__NT__)
    short killer = 0;
#endif
    short xdel = 0;

#ifdef QDOS
    int errval = (int) QLERR_FLAG;
    char *pt;

    pt = *((char **) (_sys_var + 0x7c));
    minnie = ((*((short *) (pt + 0x44)) == 4) && (*((long *) (pt + 0x46)) == 0x4A534C31));

    xbaud = ((*(_sys_var + 0xA0)) & 0x7);
    domon = 0;
    *LastNam = 0;
#endif
#ifdef _DCC
    StartRexx ();
#endif
    for (i = 0; i < 256; i++)
	reversebits[i] =
	    (i & 1 ? 128 : 0) | (i & 2 ? 64 : 0) | (i & 4 ? 32 : 0) | (i & 8 ? 16 : 0) |
	    (i & 16 ? 8 : 0) | (i & 32 ? 4 : 0) | (i & 64 ? 2 : 0) | (i & 128 ? 1 : 0);


    for(i = 1; i < ac; i++)
    {
	char *p = NULL;

	if (**(av+i) == '-')
	{
	    if (*(1+*(av+i)) == 'C')
	    {
		if (*(2+*(av+i)))
		{
		    p = (2+*(av+i));
		}
		else if (*(av + 1 + i) && **(av + 1 + i) != '-')
		{
		    p = *(av+i+1);
		}
		if (p)
		{
		    char t[PATH_MAX];
		    if(access(p, R_OK) == 0)
		    {
			sprintf(t,"QFAX_DATA=%s",p);
			putenv(t);
		    }
		}
		break;
	    }
	}
    }

    GetQFax ();
    atexit (TidyUp);

    if (qd.qfaxini)
    {
	FrigArgs (&ac, &av, qd.qfaxini);
    }

    while (!err && (c = getopt (ac, av,
		"Aaq2:D:kfcpgGvl:i:r:s:z:w:RW:xe:b:m:P:S:ZX:C:")) != -1)
    {
	switch (c)
	{
            case 'a':
                anon = 0;
                break;
            case 'A':
                anon = 1;
                break;
            case '2':
                c20 = (optarg && *optarg == '.');
                fixupfdcc();
                fixupflid();
                break;
#if defined(__linux__) || defined(__NT__)
            case 'k':
                killer = 1;
                break;
#endif
            case 'X':
                xdel = atoi(optarg);
                break;
            case 'P':
                strcpy (pollfile, optarg);
                break;
	    case 'c':
                achar = 0;
                break;
            case 'p':
                prt++;
                break;
	    case 'x':
#ifdef QDOS
	    {
		extern char *_endmsg;
		_endmsg = NULL;
	    }
#endif
	    break;
	  case 'v':
              if (STDE == NULL)
              {
                  STDE = stderr;
                  setvbuf (stderr, NULL, _IOLBF, 512);
              }
              break;
	    case 'q':
		STDE = NULL;
		break;
	  case 'l':
              if (LOG == NULL)
              {
                  LOG = fopen (optarg, "a");
              }
              break;
            case 'i':
                AddUserAT (iopt + FX_USERAT, optarg);
                break;
            case 'G':
#ifndef _DCC
                Grab = 1;
#endif
            case 'g':
                base = NULL;
                act = c;
                rings = "";
                break;
            case 'r':
                act = 'r';
                base = optarg;
                rings = NULL;
                break;
            case 'R':
                act = 'r';
                base = rings = NULL;
                break;
            case 'Z':
                act = 'Z';
                base = rings = NULL;
                break;
            case 'S':
                act = 'S';
                base = optarg;
                break;
            case 'z':
                act = 'z';
                base = optarg;
                break;
            case 's':
                act = 's';
                base = optarg;
                break;
#ifdef QDOS
            case 'm':
                domon = 1;
#endif
            case 'W':
                loop = 1;
            case 'w':
                act = 'w';
                base = NULL;
                rings = optarg;
                break;
            case 'e':
#ifdef QDOS
                errval = atoi (optarg);
                break;
#endif
            case 'b':
                bopt = atoi (optarg);
                break;
            case 'D':
                strcpy(qd.serdev, optarg);
                break;
            case 'C':
                break;
            default:
            case '?':
                err = 2;
                break;
	}
    }

    prt &= 1;
    
    if(*qd.ini[0] == 0) iopt[1] = dl_1[c20];
    if(*qd.ini[1] == 0) iopt[2] = dl_2[c20];
    
#ifdef __linux__
    {
	short gz = tolower(act);
	if(gz != 'g' && gz != 'z')
	{
	    lockfile = lock_dev(qd.serdev, &lockpid);
	    if(lockfile == NULL && lockpid)
	    {
		if(killer)
		{
		    kill(lockpid, SIGKILL);
		    lockfile = lock_dev(qd.serdev, &lockpid);
		}
		else exit(0);
	    }
	}
    }
#endif
    {
	char **ap;
	int pp;
	char strg[256], *ps;
	struct tm *tm;
	time_t t;
	
	ps = strg;

	for (pp = ac, ap = av + 1; --pp; ap++)
	{
	    strcpy (ps, *ap);
	    ps += strlen (*ap);
	    *ps++ = ' ';
	}
	*ps = '\0';
	msg (0, "qfax v" FAXVERS " %s", strg);
	t = time(NULL);
	tm = localtime(&t);
	strftime(strg, sizeof(strg)-1, "Startup on %A, %d %B %Y", tm);
	msg(0, strg);
    }

#ifdef QDOS
    if (act == 's' || act == 'S')
    {
	dact = 1;
    }
    else if (act == 'r' || act == 'R')
    {
	dact = 2;
    }
    else if (act == 'w')
    {
	dact = 4;
    }
    if (qd.sflag & dact)
    {
	Iserve = (ServerInit () == 0);
	if (*qd.StartScript)
	{
	    RunScript (qd.StartScript, -1, NULL);
	    ctx = SrvPoll ("QFAX", 0, &srvmsg, &srvlen);
	    if (ctx >= 0)
	    {
		if (strncasecmp (srvmsg, "exitparam", 9) == 0)
		{
		    ParseExitP (ctx, srvmsg);
		}
		else
		{
		    SrvRply (ctx, "\007Invalid", -1);
		    free (srvmsg);
		}
	    }
	    srvmsg = NULL;
	}
    }
#endif

    do
    {
	qmode = "error";
	online = 0;
	conf = 1;
	*tsi = 0;

	if (act == 's' || act == 'S' || act == 'z')
	{
	    char *p;
	    char c = 0;

	    c = *pollfile;

	    p = *(av + optind);
	    if (act == 'S')
	    {
		ipoll = 1;
	    }
	    else
	    {
		ipoll = 0;
	    }

	    force = 0;
	    err = FaxSend (base, p);
	    if (c)
	    {
		*pollfile = c;
	    }
	}
	else if (act)
	{
	    err = receive (base, rings);
#ifndef QDOS
	    if (err == QX_ERR_255)
	    {
		loop = 0;
	    }
#endif
	}
	else
	{
	    loop = 0;
	    msg (0, "No command - see documentation please ... ");
	    break;
	}

	{
	    char txt[256];

	    sprintf (txt, "%s:%d, connect time %d secs, %s[%s] -- Class 2%s",
		     qmode, err, (online > 0) ? online : 0,
		     (conf) ? "" : "* ", tsi, (c20) ? ".0" : "");
	    if (alias)
	    {
		int n;

		n = strlen (txt);
		sprintf (txt + n, " (%s)", ttsi);
		*ttsi = 0;
	    }

	    msg (QX_FORCE, txt);
#ifdef QDOS
	    if (srvmsg && ctx >= 0)
	    {
		SrvRply (ctx, txt, strlen (txt));
		free (srvmsg);
		srvmsg = NULL;
	    }
	    else if (Iserve)
	    {
		SrvBro ("QFAX", txt, strlen (txt));
	    }
#endif
	    
	    if (act == 'G')
	    {
#ifndef QDOS
		printf ("!%d,%d,%s,%s%s\n", err, (online > 0) ? online : 0, tsi,
			qd.in_dir, LastNam);

#else
		printf ("!%d,%d,%s,%s\n", err, (online > 0) ? online : 0, tsi,
			LastNam);
#endif
		fflush(stdout);
	    }

	}
	/* Reset LSB if using forced */
	my.lsb = qd.elsb[0];

	if (*qd.apres)
	{
	    cmd (NULL, 1);
	    cmd (CheckOpt (qd.apres), 5);
	}
	if (loop)
	{
	    Delay (15);
	    faxsync (inicmd);
	    Delay (25);
	}
    }
    while (loop);

#ifdef QDOS
    if (errval != QLERR_FLAG)
    {
	err = errval;
    }
    if (err > 0)
    {
	err = 0;
    }
#else
    TidyUp();
#endif
    if(xdel)
    {
	signal (SIGABRT, SIG_DFL);
	signal (SIGTERM, SIG_DFL);
	signal (SIGINT, SIG_DFL);
	sleep(xdel);
    }
    return err;
}
