/*--------------------.
| TTy driver for qfax |
`--------------------*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <windows.h>
#include <getopt.h>

HANDLE myfd;

long tty_open (char *dev, long baud, short flag)
{
    HANDLE f;
    DCB dcb;

    if(dev)
    {
	if((f = CreateFile (dev, GENERIC_READ|GENERIC_WRITE, 0, NULL,
                OPEN_EXISTING, 0, 0)))
	{
            myfd = f;
            GetCommState(f, &dcb);
            dcb.BaudRate = baud;
            dcb.fParity = 0;
            dcb.fOutxCtsFlow = 1;
            dcb.fDtrControl = DTR_CONTROL_ENABLE;            
            dcb.fRtsControl = RTS_CONTROL_HANDSHAKE;            
            dcb.fInX = 0;
            dcb.fOutX = 0;
            dcb.ByteSize = 8;
            dcb.Parity = NOPARITY;
            SetCommState(f, &dcb);
	}
    }
    else
    {
	f = 0;
    }   
    return (long) f;
}

void tty_close(void)
{
    CloseHandle(myfd);
    myfd = (HANDLE) -1;
}

int SWrite(char *buf, int n)
{
    long nw;
    if(n < 0)
    {
	n = strlen(buf);
    }

    WriteFile(myfd, buf, n, &nw, NULL);
    return nw;
}

static char rbuf[1024];
static short ibuf = 0;
static char *tp = rbuf;

int SRead(char *buf, int req, int t)
{
    long nr=0;
    long nc = req;
    static COMMTIMEOUTS cto = {0};
    if(t < 0) t = 0;

    while(ibuf && nc)
    {
	*(buf+nr) = *tp++;
	ibuf--;
	nc--;
	nr++;
    }

    if (nc)
    {
        long rcv;
        cto.ReadTotalTimeoutConstant = t*100;    /* Total in millisecs */
        cto.ReadIntervalTimeout = 0;
        SetCommTimeouts(myfd, &cto);
        if(ReadFile(myfd, buf, nc, &rcv, NULL) != TRUE)
        {    
            nr = -1;
        }
        else
        {
            nr += rcv;
            cto.ReadTotalTimeoutConstant = 0;    /* Total in millisecs */
            cto.ReadIntervalTimeout = MAXDWORD;
            SetCommTimeouts(myfd, &cto);
            if(ReadFile(myfd, (tp = rbuf), sizeof(rbuf), &rcv, NULL) == TRUE)
            {
                 ibuf = rcv;
            }
        }
    }
    return nr;
}

#ifdef TEST
static int faxgets (char *s, int n, int t)
{
    int c = 0;
    char *p;
    char termc = '\n';
    short tlm, tc;
    short first = 1;
    int q = 0;

    p = s;
    *p = 0;
    tlm = t * 10;
    tc = 0;
    n--;

    while (n)
    {
	c = SRead (p, 1, 1);
	if (c < 0)
	{
	    q = -1;
	    break;
	}
	else if (c == 0)
	{
	    if (tc++ == tlm)
	    {
		q = -2;
		break;
	    }
	}
	else
	{
	    if (first)
	    {
		if (tolower (*p) == 'a')
		{
		    termc = '\r';
		}
		first = 0;
	    }
	    n--;
	    if (*p++ == termc)
	    {
		break;
	    }
	}
    }
    *p = 0;
    return q;
}

short verbose = 0;

static char *cmd (char *s, int t, char *resp)
{
    static char buf[128];
    int p;
    char *q;

    if (s && *s)
    {
	SWrite ("AT", -1);
	SWrite (s, -1);
	SWrite ("\r", -1);
	if(verbose) fprintf (stderr, "Modem : AT%s ... ", s);
    }

    while ((p = faxgets (buf, 128, t)) == 0)
    {
        if(verbose)
           fputs (buf, stderr);    
	if(strstr(buf, resp))
	{
	    break;
	}
    }
    return (p == 0) ? buf : "";
}

int main(int ac, char **av)
{
    int f,c;
    int baud = 9600;
    char *dev = "com1:";
    int timo = 1;
    char *p;

    while((c =getopt(ac,av,"b:d:t:v")) != EOF)
    {
	switch(c)
	{
	  case 'b':
	    baud = atoi(optarg);
	    break;
	  case 'd':
	    dev = optarg;
	    break;
	  case 't':
	    timo = atoi(optarg);
	    break;
	  case 'v':
	    verbose = 1;
	    break;
	}
    }
    if(verbose)
	fprintf(stderr, "testTTY %s at %d\n", dev, baud);
    
    f = tty_open(dev, baud, 1);
    SWrite("AT\r", 3);
    cmd("", timo, "OK");
    
    p = cmd ("i2", timo, "OK");
    fputs((*p) ? p : "fails\n", stderr);
    
    tty_close();
    return 0;
}
#endif
