#include <stdio.h>
#include <stdlib.h>

#ifdef QDOS
# include <qdos.h>
#endif

# include <sys/stat.h>

#include <string.h>
#include "alphabet.h"
#include "qheader.h"
#include "qfaxparse.h"
#include "qfv.h"
#ifndef SLACK
#define SLACK 0
#endif

static long *PagList = NULL;
static short PagLLen = 0, PageScan = 0;

static FILE * fp = NULL;

short pixwid[3] = {1728, 2048, 2432};
short pixhgt[3] = {2286, 2804, 3218};

int lasteol = 0;

extern short toskip, skipall;

#define NT_NODE 0
#define NT_TERM 1
#define NT_MUCO 2
#define NT_REMA 3
#define NT_EOL	4

typedef struct node
{
    struct node *su0, *su1;
    short free;
    short term;
    short black;
    short type;
    short runlen;
} NODE, *PNODE; 

typedef unsigned char UCHAR, *PUCHAR;

static PNODE w_root = NULL, b_root = NULL, currnode;
static short cx, cy, expected_col, eop, line_2d;
static short prev_eols, b1, b2, parsing_fill = 0, big_mu;
static short num_errs = 0, col_of_nce, bit_num = 0;
static short sf, nbl, eof, tbl, nop;

static UCHAR *bit_tbl;

static UCHAR bit_msb[8] = {0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};
static UCHAR bit_lsb[8] = {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80};

static short F_pix, F_char, F_page, F_res, F_code;
static long twr, tbr, lastCh;
static short skipbad, badlines, lastbaddie;
static short lastpg;

UCHAR *op, *refline;

static short from = 0;
extern FAXDATA qd;
short mineop = 4;
#if defined (__GNUC__) && !defined(PRTFAX)
extern UCHAR *RedMap;
UCHAR *rb;
#ifdef AMIGA
extern ShowFax(void);
void GiveItBack(void)
{
}
#endif
#endif

long faxseek(FILE *fp, long offset, int mode)
{
    long pos;
    
    fseek(fp, offset, mode);
    pos = ftell(fp);
    lastCh = pv->fbytes - pos;
    return pos;
}


short get_nce (short bp)
{
    short cu_co;

    cu_co = ((refline[bp >> 3] & bit_msb[bp & 0x07]) != 0);

    col_of_nce = !cu_co;

    if (cu_co)
	while (((refline[bp >> 3] & bit_msb[bp & 0x07]))
	       && (bp < F_pix))
	    bp++;
    else
	while ((!(refline[bp >> 3] & bit_msb[bp & 0x07]))
	       && (bp < F_pix))
	    bp++;

    return bp;
}

/*************************************************************************/

PNODE getnewnod (void)
{
    PNODE rv;
    
    rv = (PNODE) malloc (sizeof (NODE));
    memset(rv, 0, sizeof(NODE));
    
    if (rv != NULL)
    {
	rv->su0 = rv->su1 = NULL;
	rv->free = 1;
	rv->runlen = -1;
	rv->type = NT_NODE;
	return (rv);
    }
    else
    {
	Report ("out of memory\n");
	return (rv);
    }
}

void enter_code (char *codestr, short nrbits, PNODE root_to_use, short black)
{
    short sl, i;
    PNODE ptn;

    ptn = root_to_use;
    sl = (short) strlen (codestr);

    for (i = 0; i < sl; i++)
    {
	if (codestr[i] == '0')
	{
	    if (ptn->su0)
		ptn = ptn->su0;
	    else
	    {
		ptn->su0 = getnewnod ();
		ptn = ptn->su0;
	    }
	}
	else
	{
	    if (ptn->su1)
		ptn = ptn->su1;
	    else
	    {
		ptn->su1 = getnewnod ();
		ptn = ptn->su1;
	    }
	}
    }

    if (ptn->free)
    {
	ptn->free = 0;
	ptn->runlen = nrbits;
	ptn->black = black;
	if ((nrbits < 64) && (nrbits >= 0))
	{
	    ptn->term = 1;
	    ptn->type = NT_TERM;
	}
	else if (nrbits >= 64)
	{
	    ptn->term = 0;
	    ptn->type = NT_MUCO;
	}
	else if (nrbits == -2)
	{
	    ptn->term = 1;
	    ptn->type = NT_EOL;
	}
    }
    else
    {
        Warning ("ERROR: leaf for %d, %d is occupied by %d, %d",
		nrbits, black, ptn->runlen, ptn->black);
    }
}

void make_trees (void)
{
    short i;

    w_root = getnewnod ();
    b_root = getnewnod ();

    for (i = 0; i < NRUN_CODE; i++)
    {
	enter_code (white_codes[i], i, w_root, 0);
	enter_code (black_codes[i], i, b_root, 1);
    }
    for (i = 0; i < NMAKE_CODE; i++)
    {
	enter_code (mu_code_white[i], (i + 1) * 64, w_root, 0);
	enter_code (mu_code_black[i], (i + 1) * 64, b_root, 1);
    }

    enter_code (EOL, -2, w_root, 0);
    enter_code (EOL, -2, b_root, 1);
    enter_code (FILL, -3, w_root, 0);
    enter_code (FILL, -3, b_root, 1);
}

static long getnextbyte (void)
{
    long c;

    c = fgetc(fp);

    if(--lastCh <= 0)
    {
        c = -1;
    }

    if((c == -1) || (cy == F_page))
    {
        if(c == -1)
        {
            eof++;
        }
        eop = 1;
        c = -1;
    }        
    return c;
}

/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

short nextbit (void)
{
    static UCHAR adat;
    short bit;

    if (bit_num == 0)
    {
        adat = getnextbyte();
    }
    bit = ((adat & bit_tbl[bit_num++]) != 0);
    bit_num &= 0x07;

    return bit;
}

void reset_decoder (short col)
{
    if (col)
    {
	currnode = b_root;
    }
    else
    {
	currnode = w_root;
    }
    parsing_fill = 0;
}
#if defined (__GNUC__) && !defined(PRTFAX)
static UCHAR bit[4] = {0xc0, 0x30, 0xc, 0x3};

void MakeImage(void)
{
    if(rb && (cy & 1))
    {
        short j,k,np = 0;
        *rb = 0;
	for (j = 0; j < F_char; j++)
	{
	    for (k = 0; k < 4; k++)
	    {
		if (op[j] & bit[k])
		{
		    *rb |= bit_msb[np];
		}
		np++;
		if (np == 8)
		{
		    rb++;
		    *rb = 0;
		    np = 0;
		}
	    }
	}
    }
    sf++;
#ifdef AMIGA
    if(sf == pv->ysize * (pv->qh.resolution + 1))
    {
        ShowFax();
    }
#endif
}
#endif

void process_eol (void)
{
    expected_col = 0;

    if(cx && cx != F_pix)
    {
        if(skipbad)
        {
            if(cy)
            {
                Warning("bad line at %ld (%ld)", cy, cx);
		memcpy(op, op - F_char, F_char);
	    }
            lastbaddie = cy;
	}
	badlines++;
	if (!pv->qh.resolution) badlines++;
    }

    reset_decoder (0);

    if(nbl == cy)
    {
        tbl++;
    }
    if (!prev_eols)
    {
	if(nop < 2) cy++;
        if(cy < pv->buck && nop == 0)
        {
#if defined(__GNUC__) && !defined(PRTFAX)
            MakeImage();
#endif
            if (!pv->qh.resolution)
            {
	        memcpy (op+F_char, op, F_char);
	        cy++;
	        op +=F_char;
	    }
	    op +=F_char;
        }
    }

    prev_eols++;
    cx = 0;

    if (prev_eols >= mineop || cy >= F_page || cy >= pv->buck)
    {
	prev_eols = 0;
	eop = 1;
    }
    if(toskip)
    {
	nop = --toskip;
    }
}

static short setbits (short black, short len)
{
    short i;
    static short el=-1;
    
    if(cx + len > F_pix)
    {
	if(el != cy)
	{
	    el = cy;
	    ShowReport(&qd, REP_PRT, 	    
		   "%d:Length error at %d %d (%d,%d)\n", 
		   from, cy, cx, len, F_pix);
	}
	len = F_pix - cx;
    }
    
    if (black)
    {
        tbr += len;
        if(nop)
        {
            cx += len;
        }
        else
        {
	    for (i = 0; (i < len) && (cx < F_pix); i++, cx++)
	    {
                op[cx >> 3] |= bit_msb[cx & 0x07];
	    }
	}
        nbl = cy;
    }
    else
    {
	cx += len;
        twr += len;
    }
    return 0;
}

short nextstep (short bitset)
{
    PNODE nextnode;
    short rv = 0;

    from = 0;
    
    if (!parsing_fill)
    {
	if (bitset)
	{
	    nextnode = currnode->su1;
	}
	else
	{
	    nextnode = currnode->su0;
	}
    }
    else
	nextnode = NULL;

    if (nextnode != NULL)
    {
	if ((nextnode->runlen >= 0))
	{
	    if (nextnode->runlen >= 1792)
	    {			/* color not yet defined */
		big_mu = nextnode->runlen;
		setbits (expected_col, nextnode->runlen);
		reset_decoder (expected_col);
	    }
	    else if ((nextnode->black) == (expected_col))
	    {
		setbits (nextnode->black, nextnode->runlen);
		if (nextnode->runlen < 64)
		{
		    expected_col = !expected_col;
		    prev_eols = 0;
		}
		reset_decoder (expected_col);
	    }
	    else
		ShowReport(&qd, REP_PRT, "node color\n");
		/* wrong colored endnode */ ;
	}
	else if (nextnode->runlen == -2)	/* EOL */
	{
	    process_eol ();
	    rv = 1;
	}
	else
	{
	    currnode = nextnode;
	}
    }
    else
    {
	if (currnode->runlen == -3)
	{
	    parsing_fill = 1;
	    if (bitset)		/* same as EOL */
	    {
		process_eol ();
		rv = 1;
	    }
	}
	else
	{
	    num_errs++;
	    expected_col = !expected_col;
	    reset_decoder (expected_col);
	}
    }

    return rv;

}

void reset_2d_decoder (void)
{
    b1 = b2 = 0;
    expected_col = 0;

    while ((!(refline[b1 >> 3] & bit_msb[b1 & 0x07])) && (b1 < F_pix))
    {
	b1++;
    }
}

short nextstep_2d (short bit)
{
    short zeilenende = 0;

    from = 2;
    if (bit)
    {
	/* vertical mode V(0) */
	setbits (expected_col, b1 - cx);
	expected_col = !expected_col;
    }
    else if (nextbit ())
    {				/* 01... */
	/* Vertical Mode 1 links oder rechts */
	if (nextbit ())
	{			/* 011  VR(1) */
	    setbits (expected_col, (b1 + 1) - cx);
	}
	else
	{			/* 010  VL(1) */
	    setbits (expected_col, (b1 - 1) - cx);
	}
	expected_col = !expected_col;
    }
    else if (nextbit ())
    {				/* 001 */
	/* Horizontal Mode */
	short curr_col, rv1d = 0;

	reset_decoder (expected_col);
	curr_col = expected_col;
	while ((curr_col == expected_col) && (!rv1d))
	{
	    rv1d = nextstep (nextbit ());
	}
	if (rv1d)
	{
	    Warning ("ERROR |EOL in 1D |1");
	}
	curr_col = expected_col;	/* should have switched now */

	while ((curr_col == expected_col) && (!rv1d))
	{
	    rv1d = nextstep (nextbit ());
	}
	if (rv1d)
	{
	    Warning ("ERROR |EOL in 1D |2");
	}
    }
    else if (nextbit ())
    {				/* 0001 */
	/* pass mode */
	b2 = get_nce (b1);
	setbits (expected_col, b2 - cx);
    }
    else if (nextbit ())
    {				/* 00001... */
	/* Vertical Mode  */
	if (nextbit ())
	{			/* 000011  VR(2) */
	    setbits (expected_col, (b1 + 2) - cx);
	}
	else
	{			/* 000010  VL(2) */
	    setbits (expected_col, (b1 - 2) - cx);
	}
	expected_col = !expected_col;
    }
    else if (nextbit ())
    {				/* 000001... */
	/* Vertical Mode  */
	if (nextbit ())
	{			/* 0000011  VR(3) */
	    setbits (expected_col, (b1 + 3) - cx);
	}
	else
	{			/* 0000010  VL(3) */
	    setbits (expected_col, (b1 - 3) - cx);
	}
	expected_col = !expected_col;
    }
    else if (nextbit ())
    {				/* 0000001...    */
        Warning ("Uncompressed");
    }
    else
    {				/* 0000000...    */
	short i = 7;		/* es waren schon sieben Nullen */

	while ((!nextbit ()) /* && NOT EOF*/)
	    i++;
	if ((i >= 11) /* || NOT EOF*/)	/* wenn bit gesetzt oder Puffer leergelesen */
	{
	    process_eol ();
	    zeilenende = 1;
	}
#if 0
	else
	    /* unknown */ ;
#endif
    }

    if (!zeilenende)
    {
	b1 = get_nce (cx);	        /* "next changing element" */
	if (col_of_nce == expected_col)
	    b1 = get_nce (b1);
	    prev_eols = 0;
    }
    return zeilenende;
}


short decode_line (void)
{
    short bit, eoln = 0;
    lasteol = ftell(fp);
    
    if (pv->qh.coding && line_2d)
    {
	while (!eoln && !eop)
	{
	    bit = nextbit ();
	    eoln = nextstep_2d (bit);
	}
    }
    else
    {
	while (!eoln && !eop)
	{
	    bit = nextbit ();
	    eoln = nextstep (bit);
	}
    }
    
    if (pv->qh.coding && eoln && !eop)
    {
	line_2d = !nextbit ();
	refline = op - F_char;
	reset_2d_decoder ();
    }
    return 0;
}

void CheckForChange(void)
{
    long pos;
    OFFREC o;

    pv->qh.coding = F_code;
    pv->qh.resolution = F_res;
    
    if(pv->qh.offset)
    {
        pos = ftell(fp);
	faxseek(fp, pv->qh.offset, SEEK_SET);
	ShowReport(&qd, REP_PRT, 
		   "Offseek pos = %d ", pos); 
        
        while(fread(&o, 1, sizeof(o), fp) == sizeof(o))
        {
	    OffrecEndianise(&o);
            if(abs(o.off - pos) < 4)
            {
                pv->qh.coding = o.code;
                pv->qh.resolution = o.res;
		if(o.width >= 0 && o.width < 3)
		{
		    ShowReport(&qd, REP_PRT, 
			       "CP1 code = %d res = %d width = %d", 
			       o.code, o.res, o.width);
		    if(pixwid[(short)o.width] != F_pix)
		    {
			F_pix = pixwid[(short)o.width];
			F_char = F_pix >> 3;

			pv->qh.width = o.width;
			pv->buck = (pv->maxmem / F_char) - 2;
#if defined (QDOS) && !defined(PRTFAX)
			pv->F_char = F_char;
			pv->F_pix = F_pix;
			pv->F_page = pv->buck;
#endif
		    }
		}
                break;
            }
        }
        faxseek(fp, pos, SEEK_SET);
    }
}
short decode_page (void)
{
    nbl = tbl = 0;
    line_2d = 0;
    expected_col = 0;
    big_mu = 0;
    eop = 0;
    bit_num = 0;
    sf = 0;
    cy = 0;
    lastbaddie = 0;
    
    twr = tbr = 0;

    prev_eols = 0;

    reset_decoder (0);

    CheckForChange();    

    if(skipall) toskip = skipall;
    
    while (!eop)
    {
	decode_line ();
    }
    return nbl;
}
void CloseFaxFile(void)
{
    fclose(fp);
    fp = NULL;
    if(PagList)
    {
        free (PagList);
        PagList = NULL;
        PagLLen = 0;
    }
}

FILE * FindFax(short maxlen)
{
    if(b_root == NULL)
    {
        make_trees ();
    }
    if(fp)
    {
        CloseFaxFile();
    }
    fp = fopen (pv->FileName, "rb");
    if(fp <= 0)
    {
        Report ("Can't open %s", pv->FileName);
        fp = 0;
    }
    else
    {
        struct stat s;
	PageScan = 0;
	lasteol = 0;
        lastpg = -1;
	eof = 0;
        fstat (fileno(fp), &s);
        fread (&(pv->qh), 1, sizeof(QFAXHEADER), fp);
	HeaderEndianise(&pv->qh);
        if(pv->qh.title != QFAXTITLE)
        { 
            Report("Not QFAX");
            fclose (fp);
            fp = 0;
        }
        else
        {
            F_code = pv->qh.coding;
            F_res = pv->qh.resolution;

            pv->maxwid = pv->qh.width;
            if(pv->qh.version <= QFAXFOFF)
            {
                pv->qh.offset = 0;
            }
            if(pv->qh.offset)
            {
                OFFREC o;
                pv->fbytes = pv->qh.offset;
                faxseek(fp, pv->qh.offset, SEEK_SET);
                while(fread(&o, 1, sizeof(o), fp) == sizeof(o))
                {
		    OffrecEndianise(&o);
                    if(o.width > pv->maxwid)
                    {
                        pv->maxwid = o.width;
                    }
                }
                faxseek(fp, sizeof(QFAXHEADER), SEEK_SET);
            }
            else
            {
                pv->fbytes = s.st_size;
	    }
	    lastCh = pv->fbytes;
	    
            PagLLen = (pv->qh.pagemax + 4) * sizeof(long);
            if((PagList = malloc(PagLLen)))
            {
                bzero ((char *) PagList, PagLLen);
                if(pv->qh.version < QFAXNEWFMT)
                {
                    pv->qh.page = 0;
                }

                if (pv->qh.page == 2)
                {
                    if(pv->qh.version > QFAXNEWFMT && pv->qh.hint)
                    {
                        short actual;
                        short i;
			actual = SLACK + (pv->qh.hint << (1-pv->qh.resolution));
			ShowReport(&qd, REP_PRT, 
			   "DECODE: actual = %d hint = %d res = %d %d\n",
				   actual, pv->qh.hint, pv->qh.resolution,
				   pv->fbytes );
                        pv->qh.page = 0;
                        for(i = 0; i <= maxlen; i++)
                        {
                            if(actual < pixhgt[i])
                            {
                                pv->qh.page = i;
                                break;
                            }
                        }
                        if(i >= 2)
                        {
                            pv->qh.page = pv->qh.width;
                        }
                    }
                    else
                    {
			pv->buck = 0x7fff;
                        pv->qh.page = pv->qh.width;
                    }
                }

                F_page = pixhgt[pv->qh.page];
                F_pix = pixwid[pv->qh.width];
                F_char = F_pix >> 3;

            }
            else
            {
                Report("Can't get 1K");
                fclose (fp);
                fp = 0;
	    }
            if(pv->qh.version > QFAXFLSB)
            {
                if(pv->qh.lsb)
                {
                    bit_tbl = bit_lsb;
                }
                else
                {
                    bit_tbl = bit_msb;
                }
            }
            else
            {
		pv->qh.lsb = 0;
                bit_tbl = bit_msb;
            }
        }
    }
    return fp;
}

/* Setting for LSB, swap names for MSB :-) */

static char lztail[256] =
{
    8,7,6,6,5,5,5,5,4,4,4,4,4,4,4,4,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
    1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
    1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
};

static char lzlead[256] =
{
    8,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
    5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
    6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
    5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
    7,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
    5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
    6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
    5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
};


/*
 * The reason for the static and rather weird way of doing this is
 * because this is usually in a different module
 */

static char *ztail = lztail;
static char *zlead = lzlead;

static  short zt,  nz;
static int nline;

void ZeroRTC (short flag)
{
    zt = nz = nline = 0;

    if(flag & 1)
    {
        ztail = lztail;
        zlead = lzlead;
    }
    else
    {
        zlead = lztail;
        ztail = lzlead;
    }
}
    
int TestRTC (unsigned char c, int *zzz)
{
    int rval = 0;

    nz =  nz + zlead[c];
    if (c)
    {
	if ( nz > 10)
	{
            zt++;
            nline++;
	}
	else
	{
	    if (zt >= mineop)
	    {
		if(zzz) *zzz = zt;
		nline -= zt;
		rval = nline;
	    }
	    zt = 0;
	}
        nz = ztail[c];
    }
    return rval;
}

int EndRTC(int *zzz)
{
    *zzz = zt;
    return nline;
}

void ScanFaxFile (void)
{
    int i,j;
    unsigned char buf[512];
    int c;
    int nb = 0;
    int nx = 0;
    int nc = 0;
    int np = 1;
    
    if(PageScan == 0)
    {
	PageScan = 1;
	
	ZeroRTC(pv->qh.lsb);
    
	faxseek (fp, sizeof(QFAXHEADER), SEEK_SET);

	nc = PagList[0] = sizeof(QFAXHEADER);
	
	while((j = fread(buf, 1, 512, fp)) > 0)
	{
	    for(i = 0; i < j  ; i++)
	    {
		c = buf[i];
		nx = TestRTC(c, &nb);
		if(nx && nb > mineop-1)
		{
		    PagList[np++] = nc;
		    ZeroRTC(pv->qh.lsb);
		    nx = 0;
		}
		nc++;
	    }
	}
	if(nx == 0)
	{
	    nx = EndRTC(&nb);
	    if(nb > 1)
	    {
		PagList[np++] = ftell(fp);
	    }
	}
	ShowReport(&qd, REP_PRT, "setting max, np, eof %d %d %d",
		       pv->qh.pagemax, np, eof);
    }
}

int Check4Scan (int pag)
{
    return PagList[pag];
}

int decode_fax2 (int *maxln, int page, short skipflag, short *pbad)
{
    int rv = -1;
    short uppage = 0;
    skipbad = skipflag;
    badlines = 0;
    nop = 0;

    
    if(pv->scan == NULL)
    {
        Report("No Memory");
    }

    if(fp == NULL)
    {
        return -1;
    }

    if(pv->scan != NULL)
    {
        long cp;

	if(PageScan == 0)
	{
#if defined (__linux__) || defined (__NT__)
	    ScanFaxFile();
#else
	    if(qd.prescan[0])
	    {
		ScanFaxFile();
	    }
#endif
	}
	
        do
        {
            op = (UCHAR *) pv->scan;
#if defined (__GNUC__) && !defined(PRTFAX)
            rb = RedMap;
#endif

            if(PagList[page])
            {
                faxseek(fp, PagList[page], SEEK_SET);
            }
            else
            {
                PagList[page] = ftell(fp);
            }
            decode_page();
	    if(!eof && (cy < 20 || ((tbr != 0) && twr/tbr > 1000)))
	    {
                PagList[page] = 0;
	    }
	} while (PagList[page] == 0);

        ShowReport(&qd, REP_PRT,
		   "Total %ld, black %ld, last black %ld %ld %ld\n",
		   cy, tbl, nbl, twr, tbr);

        pv->xline = nbl;
	if(maxln)
	    *maxln = cy;
        rv = 0;
        cp = ftell(fp);

        if(pv->fbytes - cp > (1024 << pv->qh.resolution))
        {
            uppage = 1;
        }
        else
        {
            uppage = 0;
            if(cp < pv->fbytes)
            {
                nop = 1;
                decode_page();
                if(tbr)
                {
                    uppage = !eof && cy > 20 && tbr && twr/tbr < 1000;
                }
                faxseek(fp, cp, SEEK_SET);
            }
            else
            {
                eof++;
            }
        }

	ShowReport(&qd, REP_PRT,
		   "uppage : cy %ld, tbl %ld, up %ld %ld %ld %ld %ld\n",
		   cy, tbl, uppage,cp , pv->fbytes, twr, tbr);

        if(eof == 0 && page > lastpg)
        {
            if(uppage)
            {
                if(page + 2 > pv->qh.pagemax)
                {
                    pv->qh.pagemax = page + 2;
	        }
            }
            else if(cy == 0)
            {
                pv->qh.pagemax = page + 1;
            }
	    lastpg = page;
        }
    }

    if(eof == 1)         /* Must be EOF and last page */
    {
        pv->qh.pagemax = page + 1;
	eof++;
    }

    *pbad = badlines;
#if 0
    printf("RSize = %d %d %d\n", (int) rb- (int)RedMap,
	   *maxln, pv->xline);
#endif
    return rv;
}
