#include <stdio.h>
#include <qdos.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>

#include "qfv.h"

Prototype void FaxInfo (struct WM_wwork *, short);

static long ACHOICE (struct WM_wwork *, struct WM_litm *, struct WM_wstat *);
static struct WM_action achoice = {JSR, wm_actli, ACHOICE};


#define VR_NORMAL	0
#define VR_FINE		1

#define DF_1DHUFFMAN 	0
#define DF_2DMREAD	1
#define DF_2DUNCOMP	2

static char *vr_string (short vr)
{
    switch (vr)
    {
	case VR_NORMAL:
	    return ("normal");
	case VR_FINE:
	    return ("fine");
	default:
	    return ("unknown");
    }
}

static char *df_string (short df)
{
    switch (df)
    {
	case DF_1DHUFFMAN:
	    return ("1D huffman");
	case DF_2DMREAD:
	    return ("2D read");
	case DF_2DUNCOMP:
	    return ("2D uncompressed");
	default:
	    return ("unknown");
    }
}

static void ioprintf(int chan, char *ctl,...)
{
    va_list va;
    char t[80];

    va_start (va, ctl);
    vsprintf (t, ctl, va);
    va_end (va);

    io_sstrg(chan, -1, t, (short)strlen(t));
}

void FaxInfo (struct WM_wwork *wwk, short badlines)
{
    static QD_TEXTI (AOK, "OK");

    static struct WM_wstat ws = {0};
    static struct WM_litm litm[1 + 1] =
    {
	{20, 14, 90, 110, 0, 0, TYP_TEXT, 3, &AOK, 0, &achoice},
	{-1}
    };

    static struct WM_infw infw[1 + 1] =	/* information window */
    {
	{396, 128, 2, 1, 0, 0, 0, 0, NULL},
	{-1}			/* end of array */
    };

    static struct WM_wwork ww =	/* window working definition */
    {
	NULL,			/* *wstat;              pointer to status area */
	NULL,			/* *wscale              this was pointer to WDEF */
	0,			/* chid                 channel ID */
	NULL,			/* *pprec               pointer to pointer record */
	0,			/* psave                pointer position saved */
	0, 0, 0, 0,		/* 4 zero */
	NULL,			/* *splst               pointer to sprite list */
	400, 130,		/* xsize, ysize */
	0, 0,			/* xorg, yorg           and origin */
	2,			/* flag                 no clear + shadow */
	1, 7, 0,		/* borw, borc, papr     border width/clr, paper */
	NULL,			/* *sprite              pointer to sprite */
	1, CL_MHIGH,		/* curw, curc           current item width/clr */
	CL_MPUNAV, CL_MIUNAV, NULL, NULL,	/* uback, uink, *ublob, *upatt unavail */
	7, 0, NULL, NULL,	/* aback, aink, *ablob, *apatt avail */
	CL_MPSLCT, CL_MISLCT, NULL, NULL,	/* sback, sink, *sblob, *spatt selected */
	NULL,			/* *help */
	1, 0, infw,		/* ninfo, ninob, *pinfo info winds */
	1, litm,		/* nlitm, *plitm        loose items */
	0, NULL			/* nappl  *pappl        appl winds */
    };

    int i;
    short xp, yp;
    
    ws.wwork = &ww;		/* cross references */
    ww.wstat = &ws;
    ws.wmode = FORM_QL4;

    ws.citem = -1;		/* no current item  */
    ws.ciact = NULL;		/* or current item routine */
    ws.litem[0] = WSI_MKAV;	/* preset ESC return */
    ww.xorg = ww.yorg = 0;

    if(wwk->xsize < 412)
    {
        ww.xsize = wwk->xsize - 12;
    }
    else
    {
        ww.xsize = 400;
    }

    if(wwk->ysize < 138)
    {
        ww.ysize = wwk->ysize - 8;
    }
    else
    {
        ww.ysize = 130;
    }

    infw[0].xsize = ww.xsize - 4;
    infw[0].ysize = ww.ysize - 2;

    litm[0].xorg = (infw[0].xsize - 20)/2;
    litm[0].yorg = infw[0].ysize - 18;
    
    xp = (wwk->xsize - ww.xsize)/2 + wwk->xorg;
    yp = (wwk->ysize - ww.ysize)/2 + wwk->yorg;

    wm_pulld (&ww, xp, yp);
    wm_wdraw (&ww);
        
    wm_swinf (&ww, 0, -1);
    sd_setst (ww.chid, -1, 0);
    sd_setpa (ww.chid, -1, 0);
    sd_setin (ww.chid, -1, 7);

    if(pv->xline)
    {
        short lpg, xpos;
        if(pv->qh.pagemax < pv->FaxPage+1)
        {
            lpg =  pv->FaxPage+1;
        }
        else
        {
            lpg = pv->qh.pagemax;
        }

        xpos = (ww.xsize - 72)/2;

        sd_pixp(ww.chid, -1, xpos, 4);
        ioprintf (ww.chid, "qfv Fax Info\n");
        ioprintf (ww.chid, "FileName   : %s\n",
                    pv->FileName);
        ioprintf (ww.chid, "TSI/CSI    : %.23s\n",
                    pv->qh.tsi);
        ioprintf (ww.chid, "Date       : %s",
                    ctime((long *)&(pv->qh.stamp)));
        ioprintf (ww.chid, "Resolution : %s\n",
                    vr_string(pv->qh.resolution));
        ioprintf (ww.chid, "Coding     : %s\n",
                    df_string(pv->qh.coding));
        ioprintf (ww.chid, "Lines      : %4d in %d secs (%d bad)\n",
                    pv->xline - BLANK, pv->eon, badlines);
        ioprintf (ww.chid, "\n                   Page %d of %d\n",
                    pv->FaxPage+1, lpg); 
    }
    else
    {
        ioprintf (ww.chid, "No valid QFAX loaded");
    }
    wm_ldraw (&ww, -1);
    while (!wm_rptr (&ww))
    {
        if (ws.evnt & PT_CAN)
    	    break;
    }

    wm_cluns (&ww);		/* throw it away */
}

static long ACHOICE (struct WM_wwork *wwk, struct WM_litm *li, struct WM_wstat *wst)
{
    return PT__CAN;		/* and return CAN event */
}
