#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <getopt.h>

#define PACK  __attribute__ ((packed))

#define DEF_LINESPACE_T  4
#define DEF_LINESPACE_B  2

typedef struct 
{
    ulong succ PACK;
    ulong pred PACK;
    unchar type PACK;
    unchar  pri PACK;
    ulong name PACK;
} AMINODE;

typedef struct
{
    AMINODE a PACK;
    ulong   rp PACK;
    ushort  res PACK;
    ushort  tf_YSize PACK; /* font height	  |    order to best */
    unchar  tf_Style PACK; /* font style	  |    match a font */
    unchar  tf_Flags PACK; /* preferences and flags  /    request. */
    ushort  tf_XSize PACK;	/* nominal font width */
    ushort  tf_Baseline PACK;	/* distance from the top of char to baseline */
    ushort  tf_BoldSmear PACK;	/* smear to affect a bold enhancement */
    ushort  tf_Accessors PACK;	/* access count */
    unchar  tf_LoChar PACK;	/* the first character described here */
    unchar  tf_HiChar PACK;	/* the last character described here */
    ulong   tf_CharData PACK;	/* the bit character data */
    ushort  tf_Modulo PACK;	/* the row modulo for the strike font data */
    ulong   tf_CharLoc PACK;	/* ptr to location data for the strike font */
				/*   2 words: bit offset then size */
    ulong   tf_CharSpace PACK;	/* ptr to words of proportional spacing data */
    ulong   tf_CharKern PACK;	/* ptr to words of kerning data */
} TFONT;

typedef struct 
{
    AMINODE dfh_DF PACK;             /* node to link disk fonts */
    ushort  dfh_FileID PACK;         /* DFH_ID */
    ushort  dfh_Revision PACK;       /* the font revision */
    ulong   dfh_Segment PACK;        /* the segment address when loaded */
    char    dfh_Name[32] PACK;
    TFONT   dfh_TF PACK;
} DFHEAD;

typedef struct 
{
    ushort bstart;
    ushort bwide;
} FLOC;

typedef struct
{
    char flag[8];
    unchar xpix;
    unchar ypix;
    unchar first;
    unchar last;
    unchar top_l;
    unchar bot_l;
    unchar right;
    unchar mode;
    long  fsize;
    u_char font[4];
} QFONT;



static ushort SH (ushort val)
{
    return (ushort) (val << 8) + (val >> 8);
}
static ulong LG (ulong val)
{
    return (ulong) (((ulong) SH (val & 0xFFFF) << 16) |
		    (ulong) SH(val >> 16));
}

static unchar mask[] = { 0x80, 0x40, 0x20, 0x10, 8, 4, 2, 1 };

static void DoubleLine (unchar **pqx, short *pqb, short t, unchar *sx,short sb)
{
    unchar *nx,*xx;
    short nb,xb;
    short i;
    
    nx = *pqx;
    nb = *pqb;
    xx = sx;
    xb = sb;
        
    for(i = 0; i < t; i++)
    {
	if(nb == 8)
	{
	    nx++;
	    nb = 0;
	}
	if(xb == 8)
	{
	    xx++;
	    xb = 0;
	}
	if(*xx & mask[xb])
	{
	    *nx |= mask[nb];
	}
	nb++;
	xb++;
    }
    *pqx = nx;
    *pqb = nb;
}

int main(int ac, char **av)
{
    FILE *fp = NULL;
    ushort *h,g;
    short i,j;
    DFHEAD *d;
    unchar *fd;
    struct stat s;
    FLOC *fl;
    short *fs, *fk;
    unchar *fx, *qx, *px;
    QFONT *qfont;
    unchar *prop;
    int avx = 0;
    long pbyte;
    short qb,pb,jx;
    int c;
    char fnam[256];
    short verbose = 0, large = 0, skip = 0;
    
    fnam[0] = 0;
    
    while((c = getopt(ac,av,"xdv")) != EOF)
    {
	switch(c)
	{
	  case 'd':
	    large = 1;
	    break;
	  case 'x':
	    skip = 1;
	    break;
	  case 'v':
	    verbose = 1;
	    break;
	}
    }
    
    if(av[optind] && *av[optind])
    {
	fp = fopen(av[optind], "r");
    }
    
    if(fp == NULL)
    {
	perror(av[optind]);
	exit(0);
    }

    if(optind+1 < ac)
    {
	strcpy(fnam, av[optind+1]);
    }
    
    
    fstat(fileno(fp), &s);
    fd = calloc(s.st_size, 1);
    fread(fd, s.st_size, 1, fp);
    fclose(fp);
    
    fx = (unchar *)fd;
    
    while(1)
    {
	g = *(short *)fx;
	g = SH(g);
	if(g == 0x4e75)
	{
	    fx += 2;
	    break;
	}	    
	fx++;
    }

    h = (ushort *)fx;
    d = (DFHEAD *)h;

    if(verbose)
    {
	printf("ID %x\n", ac, SH(d->dfh_FileID));
	printf("YS %d\n", SH(d->dfh_TF.tf_YSize));
	printf("XS %d\n", SH(d->dfh_TF.tf_XSize));
	printf("LC %d\n", d->dfh_TF.tf_LoChar);
	printf("HC %d\n", d->dfh_TF.tf_HiChar);
	printf("MO %d\n", SH(d->dfh_TF.tf_Modulo));
	printf("Data  %d\n", LG(d->dfh_TF.tf_CharData));
	printf("Prop  %d\n", LG(d->dfh_TF.tf_CharLoc));    
	printf("Space %d\n", LG(d->dfh_TF.tf_CharSpace));    
	printf("Kern  %d\n", LG(d->dfh_TF.tf_CharKern));    
    }
    
    fl = (FLOC *)   ((char *)h + LG(d->dfh_TF.tf_CharLoc)-4);
    fk = (short *)  ((char *)h + LG(d->dfh_TF.tf_CharKern)-4);
    fs = (short *)  ((char *)h + LG(d->dfh_TF.tf_CharSpace)-4);
    fx = ((unchar *)h + LG(d->dfh_TF.tf_CharData)-4);

    if(skip) exit(0);

    if(d->dfh_TF.tf_LoChar > 33)
    {
	fputs("!!!Too few chars !!!!\n", stderr);
	exit(0);
    }

    qfont = calloc(64*1024, 1);
    prop = calloc(256, 1);

    memcpy (qfont->flag, "QfaxFont", 8);
    qfont->xpix = 0;
    qfont->ypix = SH(d->dfh_TF.tf_YSize);
    qfont->top_l = DEF_LINESPACE_T;
    qfont->bot_l = DEF_LINESPACE_B;
    qfont->right = 0;
    qfont->first = d->dfh_TF.tf_LoChar;
    qfont->last = d->dfh_TF.tf_HiChar;
    qfont->fsize = 0;
    qfont->mode = 'P';
    qx = qfont->font;
    qb = 0;

    if(d->dfh_TF.tf_LoChar == 33)
    {
	short qxx;

	puts(" ******* Adding SPACE *********");
	qfont->first = 32;
	prop[0] = SH(d->dfh_TF.tf_XSize);
	qxx =  prop[0] * qfont->ypix;
	qb = qxx & 7;
	qxx = qxx >> 3;
	qx += qxx;
	jx = 1;
    }
    else
    {
	jx = 0;
    }
    
    
    for(j = 0, i = d->dfh_TF.tf_LoChar; i < d->dfh_TF.tf_HiChar+1;i++,j++)
    {
	ushort boff,wid;
	short c1,c2,t,k;
	short m,l,n;
	unchar *sx;
	short sb;
	
	boff = SH((fl+j)->bstart);
	wid = SH((fl + j)->bwide);
	c1 = SH(fs[j]);
	c2 = SH(fk[j]);

	if(c2 < 0) c2 = 0;
	
	t = c1 + c2;
	k = c1 - wid;

	if(large)
	{
	    k *= 2;
	    c2 *= 2;
	    t *= 2;
	}


	for(l = 0; l < qfont->ypix; l++)
	{
	    px = fx + l * SH(d->dfh_TF.tf_Modulo);

	    pbyte = boff;
	    pb = pbyte & 7;
	    pbyte = pbyte >> 3;
	    px = px + pbyte;

	    sx = qx;
	    sb = qb;

	    for(n = m = 0; m < k; m++)
	    {
		if(qb == 8)
		{
		    qb = 0;
		    qx++;
		}
		qb++;
		n++;
	    }
	    
	    for(m = 0; m < wid; m++)
	    {
		short res;
		
		if(pb == 8)
		{
		    pb = 0;
		    px++;
		}
		if(qb == 8)
		{
		    qb = 0;
		    qx++;
		}

		if(*px & mask[pb])
		{
		    *qx |= mask[qb];
		}

		if(large)
		{
		    qb++;
		    if(qb == 8)
		    {
			qb = 0;
			qx++;
		    }
		    if(*px & mask[pb])
		    {
			*qx |= mask[qb];
		    }
		    n++;
		}
		qb++;
		pb++;
		n++;
	    } 
	    
	    for(m = 0; m < c2; m++)
	    {
		if(qb == 8)
		{
		    qb = 0;
		    qx++;
		}
		qb++;
		n++;
	    }
	    if(large)
	    {
		DoubleLine(&qx, &qb, n, sx, sb);
	    }
	} 

	prop[jx] = n;
	jx++;
	avx += n;
    }
    
    if(large)
    {
	qfont->ypix *= 2;
    }

    if(fnam[0] == 0)
    {
	sprintf(fnam, "%s.font", av[optind]);
    }

    qfont->xpix = avx/jx;
    qfont->fsize = (int) (qx - qfont->font) + ((qb + 7) >> 3);
    pbyte = qfont->fsize + sizeof(QFONT) - 4;

    fprintf(stderr,"%s %2dx%2d (%d - %d) %d\n",
	    fnam, qfont->xpix, qfont->ypix, 
	    qfont->first, qfont->last,
	    pbyte);

    fp = fopen(fnam, "w");
    qfont->fsize = LG(qfont->fsize);
    fwrite(qfont, pbyte, 1, fp);
    fwrite(prop, jx, 1, fp);
    fclose(fp);

    return 0;
}

