#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>

#include <X11/Intrinsic.h>
#include "forms.h"

#include "xfax.h"
#include "qfv.h"
#include "xfax_p.h"

static VIEWPORT vp = {0};
PVIEWPORT pv = &vp;

#define VR_NORMAL	0
#define VR_FINE		1

#define DF_1DHUFFMAN 	0
#define DF_2DMREAD	1
#define DF_2DUNCOMP	2

FAXDATA qd =  {0};

short toskip = 0;
short skipall = 0;
char *RedMap = NULL;

static short SuperW=WIDTH_SUPER, SuperH = HEIGHT_SUPER;

static short badlines = 0;
static short skipbad = 1;

static short fact = M_LOAD|M_QUIT;

void Report (char *ctl,...)
{
#if 0
    time_t t;
    va_list va;
    va_start (va, ctl);
    t = time(NULL);
    fprintf(stderr, "%8.8s ", (ctime(&t) + 11));
    vfprintf (stderr, ctl, va);
    fputc('\n', stderr);
    va_end (va);
#endif
}

void Warning (char *ctl,...)
{
#if 0
    time_t t;
    va_list va;
    va_start (va, ctl);
    
    t = time(NULL);
    fprintf(stderr, "%8.8s ", (ctime(&t) + 11));
    vfprintf (stderr, ctl, va);
    fputc('\n', stderr);
    
    va_end (va);
#endif
}


static short CheckFile(void)
{
    short can_do = 0;
    if(pv->FaxPage > 0)
    {
        can_do |= M_PPAGE;
    }
    if(pv->FaxPage < pv->qh.pagemax - 1)
    {
        can_do |= M_NPAGE;
    }    
    return can_do;
}


static char *vr_string (short vr)
{
    switch (vr)
    {
	case VR_NORMAL:
	    return (" nml");
	case VR_FINE:
	    return (" fine");
	default:
	    return ("?");
    }
}

static char *df_string (short df)
{
    switch (df)
    {
	case DF_1DHUFFMAN:
	    return ("1D");
	case DF_2DMREAD:
	    return ("2D");
	case DF_2DUNCOMP:
	    return ("2Du");
	default:
	    return ("??");
    }
}

extern void DirNameCheck(char *);
extern short GetNameOpt(void);

static     int nline;

void show_info(void)
{
    short nlin, xlin,lsb;
    char txt[1024];
    char *p[3];

    lsb = 0;
    if(pv->qh.version > QFAXFLSB)
    {
	lsb = pv->qh.lsb;
    }
    
    nlin = pv->xline;
    xlin = pixhgt[pv->qh.page];

    p[0] = txt;
    sprintf(p[0], "%s %.23s\n", pv->FileName, pv->qh.tsi);

    p[1] = p[0] + strlen(p[0])+1;
    sprintf(p[1], "%24.24s %s%s (%csb)\n",
                    ctime((time_t*)&pv->qh.stamp),
                    df_string(pv->qh.coding), 
                    vr_string(pv->qh.resolution), 
		    lsb ? 'l' : 'm');

    p[2] = p[1] + strlen(p[1])+1;

    sprintf(p[2], "%4d:%4d (%d bad) %d/%d",
                    nlin, xlin, badlines, 1+pv->FaxPage, pv->qh.pagemax);

    fl_show_message(p[0], p[1], p[2]);
}

void LoadFax(FD_fax *f, BITMAP *b, short lflag)
{    
    static short flag = 0;
    short nopt=0;
    static XGCValues values;
    
    if(*vp.FileName)
    {
        int rv = -1;
	b->cury = b->curx = 0;        
        if(flag)
        {
	    bzero(b->data[0], b->size[0]);
	    if(b->data[1])
	    {
		bzero(b->data[1], b->size[1]);
	    }
        }

        if (lflag)
        {
            if(FindFax(2))
            {
		XFillRectangle(fl_get_display(), CANVID(f->canvas), b->sgc,
		   0, 0, b->width, b->height);

                if((pixwid[pv->qh.width] >> b->vres) != SuperW)
                {
                    b->width = SuperW = (pixwid[pv->qh.width] >> b->vres);
                }    

                if((pixhgt[pv->qh.page] >> b->vres) != SuperH)
                {
                    b->height = SuperH = (pixhgt[pv->qh.page] >> b->vres);
                }    
		b->fn = 1;
		b->xi = XCreateImage (fl_get_display(),
				      fl_state[fl_get_vclass()].xvinfo->visual,
				      1, XYBitmap, 0, 
				      b->data[b->vres], b->width, b->height, 8, 
				      b->width >> 3);
		XInitImage (b->xi);
		b->xi->bitmap_bit_order = MSBFirst;
		b->xi->byte_order = MSBFirst;
		b->pixmap = XCreatePixmap (fl_get_display(), 
					   CANVID(f->canvas),
					   b->width, b->height, 1);

	    values.foreground = 0;
	    values.background = 1;
		b->pgc = 
/*XCreateGC (fl_get_display(), b->pixmap, 0, 0);*/

		    XCreateGC (fl_get_display (), b->pixmap,
			       GCForeground | GCBackground, &values);

                if(pv->qh.version > QFAXNEWFMT && pv->qh.hint)
                {
                    pv->buck = (6 + (pv->qh.hint << (1-pv->qh.resolution)));
		}
		else
		{
		    pv->buck =  HEIGHT_SUPER;		    
		}
		DirNameCheck(vp.FileName);
	    }
	    else
	    {
                /* Sort out error */
                /* Report("Fubar"); */
	    }
	}

        rv = decode_fax2 (&nline, pv->FaxPage, skipbad, &badlines);
	
        if(rv == 0)
        {
	    nopt = GetNameOpt();
            fact = M_LOAD|M_QUIT|M_PRINT|M_PRINTX|M_FLIP|nopt;
	    DrawImage(f);
	    fact |= CheckFile();                
	    ResetAll(f,b);
        }
    }
    flag++;
    
    SetMenu(f, fact);
}

void FreeBitMap (FD_fax *f, BITMAP *b, short flag)
{
    if(b->xi)
    {
	if(flag)
	{
	    bzero(b->data[0], b->size[0]);
	    bzero(b->data[1], b->size[1]);
	}
	b->xi->data = NULL;
	XDestroyImage(b->xi);
	b->xi = NULL;
	XFreePixmap(fl_get_display(), b->pixmap);
	b->pixmap = 0;
	XFreeGC(fl_get_display(), b->pgc);
	b->pgc = 0;
    }
    
}
void Invert(FD_fax *f, BITMAP *b)
{
    int n;

    n = (b->width >> 3);
    
    if(b->vres)
    {
	FlipMap (b->data[1], ((pv->xline+1) >> 1)*n);
    }
    else
    {
	FlipMap (b->data[0],  n*pv->xline);
    }
    DrawImage(f);
}

void GetBitMap (FD_fax *f, BITMAP *b, char *fnam)
{
    if(*fnam)
    {
	if(b->xi)
	{
	    FreeBitMap(f,b,1);
	}		
	strcpy(pv->FileName, fnam);
	LoadFax(f, b, 1);
	ResetAll(f,b); 
    }
    else
    {
	SetMenu(f, fact);
    }
}


static void MultiPrint (FD_fax *f, BITMAP *b)
{
    short i = 0, res = 0;
    short dodec = 0;
    short ip;
    short np;
    int maxl;
    
    ip = pv->FaxPage;

    np = 0;
    
    do
    {
	if(dodec || (pv->FaxPage != i))
	{
	    dodec = 1;
	    memset (pv->scan, '\0', b->size[0]);
	    res = decode_fax2 (&maxl, i, skipbad, &badlines);
	}
	    
	if (res == 0)
	{
	    PrintPage (np, NULL);
	}
	pv->FaxPage = i;
	np++;
	i++;
    } while (i < pv->qh.pagemax && res == 0);

    ClosePrt(0);
    
    if (pv->FaxPage != ip)
    {
	RefreshImage(f);
    }
}

void DoPrint(FD_fax *f, BITMAP *b, short flag)
{

    if(*pv->PrtName == '\0')
    {
	PRVIEW *prv;
	
	for(prv = qd.prview; prv; prv = prv->next)
	{
	    if(prv->value == pv->xpl->type)
	    {
		if(prv->file)
		{
		    strcpy(pv->PrtName, prv->file);
		}
		break;
	    }
	}
    }
    
    if(*pv->PrtName == '|')
    {
	pv->fp = popen (pv->PrtName+1, "w");
    }
    else
    {
	if (*pv->PrtName == 0)
	{
	    sprintf (pv->PrtName, 
		     "prtqfax_%s", pv->xpl->name);
	}
	pv->fp = fopen (pv->PrtName, "w");
    }
    
    if (pv->fp)
    {
	IniPrt(&qd);
	
	if (flag == MU_PRINT)
	{
	    PrintPage (0, NULL);
	    ClosePrt(0);
	}
	else
	{
	    MultiPrint (f,b);
	}
	
	if(*pv->PrtName == '|')
	{
	    pclose(pv->fp);
	}
	else
	{
	    fclose (pv->fp);
	}
    }
    else
    {
	Report ("Failed to open printer");
    }
    pv->fp = NULL;
}

static char *CheckArg (char ***pav, int *pac)
{
    char *p;
    char **av;

    av = *pav;

    if (*++(*av))
    {
	p = *av;
    }
    else if (*(av + 1) && **(av + 1) != '-')
    {
	p = *++av;
	*pac = *pac - 1;
    }
    else
    {
	p = NULL;
    }
    *pav = av;
    return p;
}

static void FixPrtIni(void)
{
    int pac = 1;
    char **sav, **pav = malloc (8);
    char *str;

    pav[0] = "x";
    pav[1] = NULL;
    str = strdup(qd.pfaxini);

    FrigArgs (&pac, &pav, str);
    sav = pav;
    while (++pav, --pac)
    {
	char *p;

	if (**pav == '-')
	{
	    switch (*++(*pav))
	    {
	      case 'P':
		if ((p = CheckArg (&pav, &pac)) != NULL)
		{
		    mineop = atoi(p);
		}
		break;
		
	      case 't':
		if ((p = CheckArg (&pav, &pac)) != NULL)
		{
		    if((isdigit(*p)))
		    {
			short t;
		    
			t = atoi (p);
			pv->xpl = XLateXPL (t, NULL);
		    }
		    else
		    {
			pv->xpl = XLateXPL(0, p);
		    }
		}

		if((pv->xpl == NULL) || (pv->xpl && (pv->xpl->type > 99)))
		{
		    pv->xpl = XLateXPL (0, NULL);
		}
		break;
		case 's':
		    if ((p = CheckArg (&pav, &pac)) != NULL)
			pv->specific = atoi (p);
		    break;
		case 'c':
		    if ((p = CheckArg (&pav, &pac)) != NULL)
			pv->compress = atoi (p);
		    break;
		case 'o':
		    if ((p = CheckArg (&pav, &pac)) != NULL)
			strcpy (pv->PrtName, p);
		    break;
	    }
	}
    }
    free(str);
}

void GetLastFile (void)
{
    FILE *fp;
    char *p, qnam[PATH_MAX];

    sprintf (qnam, "%sQFAX-LastFax", qd.in_dir);
    if ((fp = fopen (qnam, "r")) != NULL)
    {
	if ((p = 
	     strchr (fgets (pv->FileName, sizeof (pv->FileName), fp), '\n')))
	{
	    *p = 0;
	}
	fclose (fp);
    }
}

void ParseCLI(int ac, char **av)
{
    if (qd.pfaxini)
    {
	FixPrtIni();
    }

    pv->FaxPage = 0;

    while (++av, --ac)
    {
	char *p;

	if (**av == '-')
	{
	    switch (*++(*av))
	    {
	      case 'P':
		if ((p = CheckArg (&av, &ac)) != NULL)
		{
		    mineop = atoi(p);
		}
		break;

	      case 'g':
		if ((p = CheckArg (&av, &ac)) != NULL)
		{
		    int pg = atoi(p);
		    if(pg) pv->FaxPage = pg-1;
		}
		break;
		
	      case 'J':
		skipall = 1;
	      case 'j':
		if ((p = CheckArg (&av, &ac)) != NULL)
		{
		    toskip = atoi(p);
		    if(skipall) skipall = toskip;
		}
		break;
		
	      case 'b':
		skipbad = 0;
		break;

	      case 't':
		if ((p = CheckArg (&av, &ac)) != NULL)
		{
		    if(isdigit(*p))
		    {
			short t;
			t = atoi (p);
			pv->xpl = XLateXPL (t, NULL);
		    }
		    else
		    {
			pv->xpl = XLateXPL(0, p);
		    }
		}
		if(pv->xpl == NULL || (pv->xpl && pv->xpl->type > 99))
		{
		    pv->xpl = XLateXPL (0, NULL);
		}
		break;
#if 0
	      case 'W':
		warn = 1;
		break;
#endif
	      case 'o':
		if ((p = CheckArg (&av, &ac)) != NULL)
		    strcpy (pv->PrtName, p);
		break;
	      case 'l':
		GetLastFile ();
		break;
	    }
	}
	else
	{
	    strcpy (pv->FileName, *av);
	}
    }

    if(pv->xpl == NULL || (pv->xpl && pv->xpl->type > 99))
    {
	pv->xpl = XLateXPL (0, NULL);
    }
}    


int init_xfax (FD_fax *f, int argc, char **argv)
{
    BITMAP *b;

    b =  (BITMAP *) f->vdata = (BITMAP *) calloc(sizeof(BITMAP),1);

    FaxParams(&qd);
    ParseCLI(argc, argv);
    
    pv->scan = b->data[0] = calloc(MEMSIZ,1);
    b->size[0] = MEMSIZ;
    b->vres = 1;
    b->size[1] = MEMSIZ >> 2;
    RedMap = b->data[1] = calloc(b->size[1],1);
    return 0;
}
