#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <dirent.h>
#include <unistd.h>

#include "forms.h"
#include "xfax.h"

#include <X11/Intrinsic.h>

#include "qfv.h"
#include "xfax_p.h"

void ResetAll(FD_fax *f, BITMAP *b)
{
    b->curx = b->cury = 0;
    fl_set_slider_value (f->xs, 0.0);
    fl_set_slider_value (f->ys, 0.0);

    if(b->width && b->height)
    {
	fl_set_slider_size (f->xs, (double)f->canvas->w/(double)b->width);
	fl_set_slider_size (f->ys, (double)f->canvas->h/(double)b->height);
    }
    else
    {
	fl_set_slider_size (f->xs, 1.0);
	fl_set_slider_size (f->ys, 1.0);
    }
}

int selfunc(__const struct dirent *d)
{
    char *q = strrchr(d->d_name, '.');
  
    return (q) ? (strncmp(q, ".fax", 5) == 0) : 0;
}

int cmpfunc (__const __ptr_t a, __const __ptr_t b)
{
    return(strcmp( (*(__const struct dirent **)a)->d_name,
       (*(__const struct dirent **)b)->d_name) > 0);
}

int GetList(__const char *dir, struct dirent ***pd)
{
    int n;
    n = scandir(dir, pd, selfunc, cmpfunc);
    return n;
}

void FreeList(struct dirent **pd, int n)
{
    int i;

    for(i = 0; i < n; i++)
    {
	free(pd[i]);
    }
    free(pd);
}

int MatchList(struct dirent **pd, char *p, int n)
{
    int i;
    for(i = 0; i < n; i++)
    {
	if(strcmp(pd[i]->d_name, p) == 0)
	{
	    break;
	}
    }
    return (i == n) ? -1 : i;
}

static struct dirent **pd = NULL;
static int maxdir = 0;
static int currdir = -1;
static char curpath[PATH_MAX] = {0};

void DirNameCheck(char *nam)
{	    
    char *p;
    p = strrchr(nam, '/');
    if(p)
    {
	*p = 0;
	if(strcmp(nam, curpath))     /* new dir */
	{
	    if(pd) 
	    {
		FreeList(pd, maxdir);
		pd = NULL;
		maxdir = currdir = 0;
	    }
	    maxdir = GetList(nam, &pd);
	    strcpy(curpath, nam);
	}
	*p = '/';
	p = p + 1;
    }
    else
    {
	p = getcwd(NULL, 0);
	maxdir = GetList(p, &pd);
	strcpy(curpath, p);
	free(p);
	p = nam;
    }
    
    currdir = MatchList(pd, p, maxdir);
}

char * GetItemName()
{
    if(currdir >= 0 && currdir < maxdir)
    {
	return pd[currdir]->d_name;
    }
    else
    {
	return NULL;
    }
}

short GetNameOpt(void)
{
    short opt = 0;
    if(currdir != -1)
    {
	if(currdir > 0)
	{
	    opt = M_PFAX;
	}
	if(currdir < maxdir - 1)
	{
	    opt |= M_NFAX;
	}
    }
    return opt;
}

void NextPage(FD_fax *f, BITMAP *b)
{
    if(pv->FaxPage < pv->qh.pagemax - 1)
    {
	GetFaxPage(f,b,MU_NPAGE);
    }
}

void PrevPage(FD_fax *f, BITMAP *b)
{
    if(pv->FaxPage > 0)
    {
	GetFaxPage(f,b,MU_PPAGE);
    }
}

void NextFax(FD_fax *f, BITMAP *b)
{
    char name_return[PATH_MAX];
    
    if(currdir != -1 && currdir < maxdir-1)
    {
	char *p,*q;
	currdir++;
	if((p = GetItemName()))
	{
	    pv->FaxPage = 0;
	    q = stpcpy(name_return, curpath);
	    *q++ = '/';
	    strcpy(q, p);
	    GetBitMap(f, b, name_return);
	}
    }
}

void PrevFax(FD_fax *f, BITMAP *b)
{
    char name_return[PATH_MAX];
    if(currdir > 0)
    {
	char *p,*q;
	currdir--;
	if((p = GetItemName()))
	{
	    pv->FaxPage = 0;
	    q = stpcpy(name_return, curpath);
	    *q++ = '/';
	    strcpy(q, p);
	    GetBitMap(f, b, name_return);
	}
    }
}




